#include <windows.h>
#ifdef WINELIB
#include <unix.h>
#include <peexe.h>
#endif
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include "common.h"
#include "exedump.h"
#include "extrnvar.h"

#pragma pack(1)

#ifdef __LCC__
#define _strnicmp strnicmp
#endif
char *DisplayType(int type);
extern int TotalFileSize;
PIMAGE_SYMBOL   PCOFFSymbolTable = 0;   // RVA to COFF symbol table (if present)

PIMAGE_SYMBOL   pSymbolTable;
char           *stringTable;
DWORD           COFFSymbolCount = 0;    // Number of symbols in COFF symbol table
long ObjLineNumberIdx = 0;
int NumberOfSymbols;
char SourceFileName[256];
char *pResourceData;
PIMAGE_SECTION_HEADER pCodeSection, pDataSection;
PIMAGE_SECTION_HEADER pSections;
int             DoDisassemble(int idx, int len, unsigned char *rec_buf);
void DumpCodeViewTypes(PBYTE ptr, DWORD length,int checkSignature,long *lp);
typedef struct {
  WORD            flag;
  PSTR            name;
}               WORD_FLAG_DESCRIPTIONS;

typedef struct {
  DWORD           flag;
  PSTR            name;
}               DWORD_FLAG_DESCRIPTIONS;

// Bitfield values and names for the IMAGE_FILE_HEADER flags
WORD_FLAG_DESCRIPTIONS ImageFileHeaderCharacteristics[] =
{
  {IMAGE_FILE_RELOCS_STRIPPED, "RELOCS_STRIPPED"},
  {IMAGE_FILE_EXECUTABLE_IMAGE, "EXECUTABLE_IMAGE"},
  {IMAGE_FILE_LINE_NUMS_STRIPPED, "LINE_NUMS_STRIPPED"},
  {IMAGE_FILE_LOCAL_SYMS_STRIPPED, "LOCAL_SYMS_STRIPPED"},
/*{ IMAGE_FILE_MINIMAL_OBJECT, "MINIMAL_OBJECT" },*/
/*{ IMAGE_FILE_UPDATE_OBJECT, "UPDATE_OBJECT" },*/
/*{ IMAGE_FILE_16BIT_MACHINE, "16BIT_MACHINE" },*/
  {IMAGE_FILE_BYTES_REVERSED_LO, "BYTES_REVERSED_LO"},
  {IMAGE_FILE_32BIT_MACHINE, "32BIT_MACHINE"},
/*{ IMAGE_FILE_PATCH, "PATCH" },*/
  {IMAGE_FILE_SYSTEM, "SYSTEM"},
  {IMAGE_FILE_DLL, "DLL"},
  {IMAGE_FILE_BYTES_REVERSED_HI, "BYTES_REVERSED_HI"}
};

#define NUMBER_IMAGE_HEADER_FLAGS \
    (sizeof(ImageFileHeaderCharacteristics) / sizeof(WORD_FLAG_DESCRIPTIONS))

PIMAGE_FILE_HEADER pImageFileHeader;
/*
Dump the IMAGE_FILE_HEADER for a PE file or an OBJ

At the beginning of an object file, or immediately after the signature of an image
file, there is a standard COFF header of the following format. Note that the Windows
NT loader limts the Number of Sections to 96.
Offset	Size	Field Description
0		2		Machine	Number identifying type of target machine. See Section 3.3.1
				Machine Types, for more information.
2		2		Number of Sections	Number of sections; indicates size of the
				Section Table, which immediately follows the headers.
4		4		Time/Date Stamp	Time and date the file was created.
8		4		Pointer to Symbol Table	Offset, within the COFF file, of the symbol
				table.
12		4		Number of Symbols	Number of entries in the symbol table. This data
				can be used in locating the string table, which immediately follows
			    the symbol table.
16		2		Optional Header Size	Size of the optional header, which is
				included for executable files but not object files. An object file
				should have a value of 0 here. The format is described in the
				section Optional Header.
18		2		Characteristics	Flags indicating attributes of the file. See
				Section 3.3.2, Characteristics, for specific flag values.

3.3.1. Machine Types
The Machine field has one of the following values, defined below, which specify its
machine (CPU) type. An image file can be run only on the specified machine, or a
system emulating it.
Constant					Value	Description
IMAGE_FILE_MACHINE_UNKNOWN	0x0		Contents assumed to be applicable to any machine
									type.
IMAGE_FILE_MACHINE_I386		0x14c	Intel 386 or later, and compatible processors.
IMAGE_FILE_MACHINE_R4000	0x166	MIPS little endian.
IMAGE_FILE_MACHINE_ALPHA	0x184	Alpha AXP.
IMAGE_FILE_MACHINE_M68K		0x268	Motorola 68000 series.
IMAGE_FILE_MACHINE_POWERPC	0x1F0	Power PC, little endian.
IMAGE_FILE_MACHINE_PARISC	0x290	Precision Architecture (PA) RISC processor.

3.3.2. Characteristics
The Characteristics field contains flags that indicate attributes of the object or
image file. The following flags are currently defined:
Flag							Value	Description
IMAGE_FILE_RELOCS_STRIPPED		0x0001	Image only. Indicates that the file does not
										contain base relocations and must therefore
										be loaded at its preferred base address. If
										the base address is not available, the loader
									    reports an error. Operating systems running
										on top of MS-DOS (Win32s) are generally not
										able to use the preferred  base address and
										so cannot run these images. However,
										beginning with version 4.0, Windows will use
									    an applications preferred base address.
IMAGE_FILE_EXECUTABLE_IMAGE		0x0002	Image only. Indicates that the image file is
										valid and can be run. If this flag is not set,
									    it generally indicates a linker error.
IMAGE_FILE_LINE_NUMS_STRIPPED	0x0004	COFF line numbers have been removed.
IMAGE_FILE_LOCAL_SYMS_STRIPPED	0x0008	COFF symbol table entries for local symbols
										have been removed.
IMAGE_FILE_MINIMAL_OBJECT		0x0010	Reserved for future use.
IMAGE_FILE_UPDATE_OBJECT		0x0020	Reserved for future use.
IMAGE_FILE_16BIT_MACHINE		0x0040	Use of this flag is reserved for future use.
IMAGE_FILE_BYTES_REVERSED_LO	0x0080	Little endian: LSB precedes MSB in memory.
IMAGE_FILE_32BIT_MACHINE		0x0100	Machine based on 32-bit-word architecture.
IMAGE_FILE_DEBUG_STRIPPED		0x0200	Debugging information removed from image file.
IMAGE_FILE_PATCH				0x0400	Reserved for future use.
IMAGE_FILE_SYSTEM				0x1000	The image file is a system file, not a user program.
IMAGE_FILE_DLL					0x2000	The image file is a dynamic-link library (DLL).
										Such files are considered executable files for
										almost all purposes, although they cannot be
										directly run.
IMAGE_FILE_BYTES_REVERSED_HI	0x8000	Big endian: MSB precedes LSB in memory.
*/
void            DumpHeader(PIMAGE_FILE_HEADER pi)
{
  UINT            headerFieldWidth = 30;
  UINT            i, stringtableOffset;
  char           *szMachine, *p, *q;
  
  pImageFileHeader = pi;
  stringtableOffset = pImageFileHeader->PointerToSymbolTable +
    (pImageFileHeader->NumberOfSymbols * 18);
  NumberOfSymbols = pImageFileHeader->NumberOfSymbols;
  if (fExtract) return;
  printf("File Header\n");
  switch (pImageFileHeader->Machine) {
  case IMAGE_FILE_MACHINE_I386:
    szMachine = "i386";
    break;
    /*      case IMAGE_FILE_MACHINE_I860:   szMachine = "i860"; break;*/
  case IMAGE_FILE_MACHINE_R3000:
    szMachine = "R3000";
    break;
  case IMAGE_FILE_MACHINE_R4000:
    szMachine = "R4000";
    break;
  case IMAGE_FILE_MACHINE_ALPHA:
    szMachine = "alpha";
    break;
  default:
    szMachine = "unknown";
    break;
  }
  
  printf("  %-*s%04X (%s)\n", headerFieldWidth, "Machine:",
	 pImageFileHeader->Machine, szMachine);
  printf("  %-*s%d\n", headerFieldWidth, "Number of Sections:",
	 pImageFileHeader->NumberOfSections);
  printf("  %-*s%08lX ", headerFieldWidth, "TimeDateStamp:",
	 pImageFileHeader->TimeDateStamp);
  q = p = ctime(&pImageFileHeader->TimeDateStamp);
  if (p == NULL) {
    printf("Wrong time field %ld\n",pImageFileHeader->TimeDateStamp);
  }
  else {
    p += strlen(p) - 2;
    while (*p == ' ')
      p--;
    p++;
    *p = 0;
    printf(" (%s)\n", q);
  }
  printf("  %-*s%08ld\n", headerFieldWidth, "PointerToSymbolTable:",
	 pImageFileHeader->PointerToSymbolTable);
  printf("  %-*s%08d\n", headerFieldWidth, "NumberOfSymbols:",NumberOfSymbols);
  
  if (pImageFileHeader->PointerToSymbolTable)
    printf("  %-*s%d\n", headerFieldWidth, "String Table File Offset",
	   stringtableOffset);
  printf("  %-*s%04X\n", headerFieldWidth, "SizeOfOptionalHeader:",
	 pImageFileHeader->SizeOfOptionalHeader);
  printf("  %-*s%04X\n", headerFieldWidth, "Characteristics:",
	 pImageFileHeader->Characteristics);
  for (i = 0; i < NUMBER_IMAGE_HEADER_FLAGS; i++) {
    if (pImageFileHeader->Characteristics &
	ImageFileHeaderCharacteristics[i].flag)
      printf("    %s\n", ImageFileHeaderCharacteristics[i].name);
  }
  
}
#if 0
// Bitfield values and names for the DllCharacteritics flags
WORD_FLAG_DESCRIPTIONS DllCharacteristics[] =
{
/*{ IMAGE_LIBRARY_PROCESS_INIT, "PROCESS_INIT" },*/
/*{ IMAGE_LIBRARY_PROCESS_TERM, "PROCESS_TERM" },*/
  {IMAGE_LIBRARY_THREAD_INIT, "THREAD_INIT"},
/*{ IMAGE_LIBRARY_THREAD_TERM, "THREAD_TERM" },*/
};
#define NUMBER_DLL_CHARACTERISTICS \
    (sizeof(DllCharacteristics) / sizeof(WORD_FLAG_DESCRIPTIONS))
// Bitfield values and names for the LoaderFlags flags
DWORD_FLAG_DESCRIPTIONS LoaderFlags[] =
{
  {IMAGE_LOADER_FLAGS_BREAK_ON_LOAD, "BREAK_ON_LOAD"},
  {IMAGE_LOADER_FLAGS_DEBUG_ON_LOAD, "DEBUG_ON_LOAD"}
};
#define NUMBER_LOADER_FLAGS \
    (sizeof(LoaderFlags) / sizeof(DWORD_FLAG_DESCRIPTIONS))
#endif

// Names of the data directory elements that are defined
char           *ImageDirectoryNames[] = {
  "EXPORT", "IMPORT", "RESOURCE", "EXCEPTION", "SECURITY", "BASERELOC",
  "DEBUG", "COPYRIGHT", "GLOBALPTR", "TLS", "LOAD_CONFIG"};

#define NUMBER_IMAGE_DIRECTORY_ENTRYS \
    (sizeof(ImageDirectoryNames)/sizeof(char *))

/*
 Dump the IMAGE_OPTIONAL_HEADER from a PE file

Every image file has an Optional Header that provides information to the loader.
This header is also referred to the PE Header. This header is optional in the sense
that some files (specifically, object files) do not have it. For image files, this
header is required. An object file may have an optional header, but generally
this header has no function in an object file except to increase size.

The Optional Header itself has three major parts:
Offset	Size	Header part	Description

0		28		Standard fields	These are defined for all implementations of COFF,
				including UNIX.
28		68		NT-specific fields	These include additional fields to support
				specific features of Windows NT (for example, subsystem).
96		128		Data directories	These fields are address/size pairs for special
				tables, found in the image file and  used by the operating system
				(for example, Import Table and Export Table).
3.4.1. Optional Header Standard Fields (Image Only)
The first nine fields of the Optional Header are standard fields, defined for every
implementation of COFF. These fields contain general information useful for loading
and running an executable file.
Offset	Size	Field	Description

0		2		Magic	Unsigned integer identifying the state of the image file.
				The most common number is 0413 octal (0x10B), identifying it as a
				normal executable file. 0407 (0x107) identifies a ROM image.
2		1		LMajor	Linker major version number.
3		1		LMinor	Linker minor version number.
4		4		Code Size	Size of the code (text) section, or the sum of all code
				sections if there are multiple sections.
8		4		Initialized Data Size	Size of the initialized data section, or the
				sum of all such sections if there are multiple data sections.
12		4		Uninitialized Data Size	Size of the uninitialized data section (BSS),
				or the sum of all such sections if there are multiple BSS sections.
16		4		Entry Point RVA	Address of entry point, relative to image base, when
				executable file is loaded into memory. For program images, this is
				the starting address. For device drivers, this is the address of the
			    initialization function.
20		4		Base Of Code	Address, realtive to image base, of beginning of
				code section, when loaded into memory.
24		4		Base Of Data	Address, realtive to image base, of beginning of
				data section, when loaded into memory.
3.4.2. Optional Header Windows NT-Specific Fields (Image Only)
The next twenty-one fields are an extension to the COFF Optional Header format and
contain additional information needed by the linker and loader in Windows NT.
Offset	Size	Field	Description

28		4		Image Base	Preferred address of first byte of image when loaded
				into memory; must be a multiple of 64K.
32		4		Section Alignment	Alignment (in bytes) of sections when loaded
				into memory. Must greater or equal to File Alignment. Default is
				the page size for the architecture.
36		4		File Alignment	Alignment factor (in bytes) used to align pages in
				image file. The value should be a power of 2 between 512 and 64K
				inclusive.
40		2		OS Major	Major version number of required OS.
42		2		OS Minor	Minor version number of required OS.
44		2		User Major	Major version number of image.
46		2		User Minor	Minor version number of image.
48		2		SubSys Major	Major version number of subsystem.
50		2		SubSys Minor	Minor version number of subsystem.

52		4		Reserved

56		4		Image Size	Size, in bytes, of image, including all headers; must
				be a multiple of Section Alignment.
60		4		Header Size	Combined size of MS-DOS Header, PE Header, and Object
				Table.
64		4		File Checksum	Image file checksum. The algorithm for computing is
				incorporated into IMAGHELP.DLL. The following are checked for
				validation at load time: all drivers, any DLL loaded at boot time,
				and any DLL that ends up in the server.
68		2		SubSystem	Subsystem required to run this image. See 'Windows NT
				Subsystem' below for more information.
70		2		DLL Flags	Obsolete.
72		4		Stack Reserve Size	Size of stack to reserve. Only the Stack Commit
				Size is committed; the rest is made available one page at a time,
				until reserve size is reached.
76		4		Stack Commit Size	Size of stack to commit.
80		4		Heap Reserve Size	Size of local heap space to reserve. Only the
				Heap Commit Size is committed; the rest is made available one page
				at a time, until reserve size is reached.
84		4		Heap Commit Size	Size of local heap space to commit.
88		4		Loader Flags	Obsolete.
92		4		Number of Data Directories	Number of data-dictionary entries in the
				remainder of the Optional Header. Each describes a location and size.
Windows NT Subsystem
The following values are defined for the Subsystem field of the Optional Header.
They determine what, if any, Windows NT subsystem is required to run the image.
Constant					Value	Description

IMAGE_SUBSYSTEM_UNKNOWN		0	Unknown subsystem.
IMAGE_SUBSYSTEM_NATIVE		1	Used for device drivers and native Windows NT
								processes.
IMAGE_SUBSYSTEM_WINDOWS_GUI	2	Image runs in the Windows graphical user interface
								(GUI) subsystem.
IMAGE_SUBSYSTEM_WINDOWS_CUI	3	Image runs in the Windows character subsystem.
IMAGE_SUBSYSTEM_POSIX_CUI	7	Image runs in the Posix character subsystem.

3.4.3. Optional Header Data Directories (Image Only)
Each data directory gives the address and size of a table or string used by Windows
NT. These are all loaded into memory so that they can be used by the system at run
time. A data directory is an eight-byte field that has the following declaration:

typedef struct _IMAGE_DATA_DIRECTORY {
    DWORD   RVA;
    DWORD   Size;
} IMAGE_DATA_DIRECTORY, *PIMAGE_DATA_DIRECTORY;

The first field, RVA, is the relative virtual address of the table. The RVA is the
address of the table, when loaded, relative to the base address of the image. The
second field gives the size in bytes. The data directories, which form the last part
of the Optional Header, are listed below.
Offset	Size	Field			Description

96		8		Export Table	Export Table address and size.
104		8		Import Table	Import Table address and size
112		8		Resource Table	Resource Table address and size.
120		8		Exception Table	Exception Table address and size.
128		8		Security Table	Security Table address and size. Unusued.
136		8		Base Relocation Table	Base Relocation Table address and size.
144		8		Debug			Debug data starting address and size.
152		8		Copyright		Copyright string address and length.
160		8		Global Ptr		Relative virtual address of the global pointer
								register. Size member of this structure is set to 0.
168		8		TLS Table		Thread Local Storage (TLS) Table address and size.
176		8		Load Config Table	Load Configuration Table address and size.

184		40		Reserved

*/
void            DumpOptionalHeader(PIMAGE_OPTIONAL_HEADER optionalHeader)
{
  UINT            width = 35;
  char           *s;
  UINT            i;
  
  printf("Optional Header\n");
  
  printf("  %-*s0x%-10x %d\n", width, "Magic", optionalHeader->Magic,optionalHeader->Magic);
  printf("  %-*s%u.%02u\n", width, "linker version",
	 optionalHeader->MajorLinkerVersion,
	 optionalHeader->MinorLinkerVersion);
  printf("  %-*s0x%-10lX %ld\n", width, "size of code", 
	 optionalHeader->SizeOfCode,
           optionalHeader->SizeOfCode);
  printf("  %-*s0x%-10lX %ld\n", width, "size of initialized data",
	 optionalHeader->SizeOfInitializedData,
	 optionalHeader->SizeOfInitializedData);
  printf("  %-*s0x%-10lX %ld\n", width, "size of uninitialized data",
	 optionalHeader->SizeOfUninitializedData,
	 optionalHeader->SizeOfUninitializedData);
  printf("  %-*s0x%-10lX %ld\n", width, "entrypoint RVA",
	 optionalHeader->AddressOfEntryPoint,
	 optionalHeader->AddressOfEntryPoint);
  printf("  %-*s0x%-10lX %ld\n", width, "base of code",
	 optionalHeader->BaseOfCode,optionalHeader->BaseOfCode);
    printf("  %-*s0x%-10lX %ld\n", width, "base of data",
			optionalHeader->BaseOfData,optionalHeader->BaseOfData);
    printf("  %-*s0x%-10lX %ld\n", width, "image base",
			optionalHeader->ImageBase,optionalHeader->ImageBase);

    printf("  %-*s0x%-10lX %ld\n", width, "section align",optionalHeader->SectionAlignment,
           optionalHeader->SectionAlignment);
    printf("  %-*s0x%-10lX %ld\n", width, "file align", 
           optionalHeader->FileAlignment,
           optionalHeader->FileAlignment);
    printf("  %-*s%u.%02u\n", width, "required OS version",
           optionalHeader->MajorOperatingSystemVersion,
           optionalHeader->MinorOperatingSystemVersion);
    printf("  %-*s%u.%02u\n", width, "image version",
           optionalHeader->MajorImageVersion,
           optionalHeader->MinorImageVersion);
    printf("  %-*s%u.%02u\n", width, "subsystem version",
           optionalHeader->MajorSubsystemVersion,
           optionalHeader->MinorSubsystemVersion);
    printf("  %-*s0x%lX\n", width, "Win32 Version", optionalHeader->Win32VersionValue);
    printf("  %-*s0x%-10lX %ld\n", width, "size of image",
			optionalHeader->SizeOfImage, optionalHeader->SizeOfImage);
    printf("  %-*s0x%-10lX %ld\n", width, "size of headers",optionalHeader->SizeOfHeaders,
           optionalHeader->SizeOfHeaders);
    printf("  %-*s0x%lX\n", width, "checksum", optionalHeader->CheckSum);
    switch (optionalHeader->Subsystem) {
    case IMAGE_SUBSYSTEM_NATIVE:
        s = "Native";
        break;
    case IMAGE_SUBSYSTEM_WINDOWS_GUI:
        s = "Windows GUI";
        break;
    case IMAGE_SUBSYSTEM_WINDOWS_CUI:
        s = "Windows character";
        break;
    case IMAGE_SUBSYSTEM_OS2_CUI:
        s = "OS/2 character";
        break;
    case IMAGE_SUBSYSTEM_POSIX_CUI:
        s = "Posix character";
        break;
    default:
        s = "unknown";
    }
    printf("  %-*s0x%X (%s)\n", width, "Subsystem",
           optionalHeader->Subsystem, s);

    printf("  %-*s0x%X\n", width, "DLL flags",
           optionalHeader->DllCharacteristics);
    if (optionalHeader->DllCharacteristics)
        printf("\n");

    printf("  %-*s0x%-10lX %ld\n", width, "stack reserve size",
           optionalHeader->SizeOfStackReserve,
           optionalHeader->SizeOfStackReserve);
    printf("  %-*s0x%-10lX %ld\n", width, "stack commit size",
           optionalHeader->SizeOfStackCommit,
           optionalHeader->SizeOfStackCommit);
    printf("  %-*s0x%-10lX %ld\n", width, "heap reserve size",
           optionalHeader->SizeOfHeapReserve,
           optionalHeader->SizeOfHeapReserve);
    printf("  %-*s0x%-10lX %ld\n", width, "heap commit size",
           optionalHeader->SizeOfHeapCommit,
           optionalHeader->SizeOfHeapCommit);

    printf("  %-*s0x%lX\n", width, "loader flags",
           optionalHeader->LoaderFlags);
    if (optionalHeader->LoaderFlags)
        printf("\n");

    printf("  %-*s0x%lX\n", width, "RVAs & sizes",
           optionalHeader->NumberOfRvaAndSizes);

    printf("\nData Directory\n");
    for (i = 0; i < optionalHeader->NumberOfRvaAndSizes; i++) {
        printf("  %-12s rva: 0x%-8lX  size: %8ld\n",
               (i >= NUMBER_IMAGE_DIRECTORY_ENTRYS)
               ? "unused" : ImageDirectoryNames[i],
               optionalHeader->DataDirectory[i].VirtualAddress,
               optionalHeader->DataDirectory[i].Size);
    }
}

// Bitfield values and names for the IMAGE_SECTION_HEADER flags
DWORD_FLAG_DESCRIPTIONS SectionCharacteristics[] =
{
    {IMAGE_SCN_CNT_CODE, "CODE"},
    {IMAGE_SCN_CNT_INITIALIZED_DATA, "INITIALIZED_DATA"},
    {IMAGE_SCN_CNT_UNINITIALIZED_DATA, "UNINITIALIZED_DATA"},
    {IMAGE_SCN_LNK_INFO, "LNK_INFO"},
    {IMAGE_SCN_LNK_REMOVE, "LNK_REMOVE"},
    {IMAGE_SCN_LNK_COMDAT, "LNK_COMDAT"},
    {IMAGE_SCN_MEM_DISCARDABLE, "MEM_DISCARDABLE"},
    {IMAGE_SCN_MEM_NOT_CACHED, "MEM_NOT_CACHED"},
    {IMAGE_SCN_MEM_NOT_PAGED, "MEM_NOT_PAGED"},
    {IMAGE_SCN_MEM_SHARED, "MEM_SHARED"},
    {IMAGE_SCN_MEM_EXECUTE, "MEM_EXECUTE"},
    {IMAGE_SCN_MEM_READ, "MEM_READ"},
    {IMAGE_SCN_MEM_WRITE, "MEM_WRITE"},
};

#define NUMBER_SECTION_CHARACTERISTICS \
    (sizeof(SectionCharacteristics) / sizeof(DWORD_FLAG_DESCRIPTIONS))

/*
Dump the section table from a PE file or an OBJ

Each row of the Section Table, is a section header. This table immediately follows
the COFF optional header, if any. This positioning is required because the file
header does not contain a direct pointer to the section table; the location of the
section table is determined by calculating the location of the first byte after the
headers.
The number of entries in the Section Table is given by the Number of Sections field
in the COFF header. Entries in the Section Table are numbered starting from one.
The code and data memory section entries are in the order chosen by the linker.
In an image file, the virtual addresses for sections must be assigned by the linker
such that they are in ascending order and adjacent, and they must be a multiple of
the Section Align value in the optional header.
Each section header (Section Table entry) has the following format, for a total of
40 bytes per entry:
Offset	Size	Field		Description

0		8		Section Name An eight-byte, null-padded ASCII string. There is no
							terminating null if the string is exactly eight bytes
							long. For longer names, this field contains an offset
							into the string table.
8		4		Virtual SizeTotal size of the section when loaded into memory. If
							this value is greater than Size of Raw Data, the section
						    is zero-padded. This field is valid only for image files
						    and should be set to 0 for object files.
12		4		RVA/Offset	Address of the first byte of the section, when loaded
							into memory, relative to the image base.For object files,
						    this field is the address of the first byte before
							relocation is applied; for simplicity, compilers should
							set this to zero. Otherwise, it is an arbitrary value
							that is subtracted from offsets during relocation.
16		4		Size of Raw Data
							Size of the section (object file) or size of the
							initialized data on disk (image files). Must be a
							multiple of File Align in the PE Header. This field is
							rounded upward to be a multiple of File Align, which is
							the only way it can be greater than Virtual Size.
20		4		Pointer to Raw Data
							File pointer to sections first page within the COFF
							file. For image files, this value must be a multiple of
							File Align in the PE Header. For object files, the value
							should be aligned on a four-byte boundary for best
							performance.
24		4		Pointer to Relocs
							File pointer to beginning of relocation entries
							for the section. Set to 0 for executable images.
28		4		Pointer to Linenumbers
							File pointer to beginning of line-number
							entries for the section.
32		2		Number of Relocs
							Number of relocation entries for the section. Set
							to 0 for executable images.
34		2		Number of Linenumbers
							Number of line-number entries for the section.
36		4		Section Flags
							Flags describing sections characteristics. See Section
							4.1, 'Section Flags,' for more information.
*/
extern LPVOID lpFileBase;
static char *PrintSectionName(char *name)
{
	static char namebuf[50];
	int i,j;

	for (i=0,j=0; i<8; i++) {
		if (name[i] >= ' ' && name[i] < 127) {
			namebuf[j++] = name[i];
		}
		else if (name[i] == 0)
			namebuf[j++] = ' ';
		else {
			namebuf[j++] = '\\';
			sprintf(namebuf+j,"%03d",name[i]);
			j += 4;
		}
		
	}
	namebuf[j] = 0;
	return namebuf;
}

void            DumpSectionTable(PIMAGE_SECTION_HEADER section,
                                 unsigned cSections,
                                 BOOL IsEXE)
{
    unsigned        i, j;

    printf("Section Table\n");
	pSections = (PIMAGE_SECTION_HEADER)(pImageFileHeader+1);

    for (i = 1; i <= cSections; i++, section++) {
		if (!fExtract) {
	        printf("  %02d %-9.9s  %s: %08ld  VirtAddr:  %08ld 0x%08lx\n",
               i, PrintSectionName(section->Name),
               IsEXE ? "VirtSize" : "PhysAddr",
               section->Misc.VirtualSize, section->VirtualAddress,section->VirtualAddress);
    	    printf("    raw data offs: %05ld raw data size: %05ld\n",
               section->PointerToRawData,
               section->SizeOfRawData);
        	printf("    relocation offs: %08ld  relocations:   %08d\n",
               section->PointerToRelocations, section->NumberOfRelocations);
        	printf("    line # offs:     %08ld  line #'s:      %08d\n",
               section->PointerToLinenumbers, section->NumberOfLinenumbers);
        	printf("    characteristics: 0x$%08lX\n", section->Characteristics);

        	printf("    ");
        	for (j = 0; j < NUMBER_SECTION_CHARACTERISTICS; j++) {
            	if (section->Characteristics &
                    SectionCharacteristics[j].flag)
                	printf("  %s", SectionCharacteristics[j].name);
        	}
        	printf("\n\n");
		}
        if (!strcmp(".text", section->Name))
            pCodeSection = section;
        if (!strcmp(".data", section->Name))
            pDataSection = section;
		if (fExtract) {
			int l = strlen(ExtractSectionName);
			if (!strncmp(section->Name,ExtractSectionName,l)) {
				char outfile[50],*p;
				FILE *f;

				p = ExtractSectionName;
				if (*p == '.') p++;
				strcpy(outfile,p);
				strcat(outfile,".sec");
				fprintf(stderr,"Extracting section %s to %s\n",section->Name,outfile);
				f = fopen(outfile,"wb");
				if (f) {
					fwrite(MakePtr(PBYTE, lpFileBase, section->PointerToRawData),
							1,(DWORD)section->SizeOfRawData,f);
					fclose(f);
				}
				else fprintf(stderr,"Impossible to write to %s\n",outfile);
			}
		}
    }
}


// Used by the DumpSymbolTable() routine.  It purpose is to filter out
// the non-normal section numbers and give them meaningful names.
//
void            GetSectionName(WORD section, PSTR buffer, unsigned cbBuffer)
{
    char            tempbuffer[10];

    switch ((SHORT) section) {
    case IMAGE_SYM_UNDEFINED:
        strcpy(tempbuffer, "UNDEF");
        break;
    case IMAGE_SYM_ABSOLUTE:
        strcpy(tempbuffer, "ABS  ");
        break;
    case IMAGE_SYM_DEBUG:
        strcpy(tempbuffer, "DEBUG");
        break;
    default:
        sprintf(tempbuffer, "%-5X", section);
    }

    strncpy(buffer, tempbuffer, cbBuffer - 1);
}

// The names of the first group of possible symbol table storage classes
char           *SzStorageClass1[] = {
    "NULL", "AUTOMATIC", "EXTERNAL", "STATIC", "REGISTER", "EXTERNAL_DEF", "LABEL",
    "UNDEFINED_LABEL", "MEMBER_OF_STRUCT", "ARGUMENT", "STRUCT_TAG",
    "MEMBER_OF_UNION", "UNION_TAG", "TYPE_DEFINITION", "UNDEFINED_STATIC",
    "ENUM_TAG", "MEMBER_OF_ENUM", "REGISTER_PARAM", "BIT_FIELD"
};

// The names of the second group of possible symbol table storage classes
char           *SzStorageClass2[] = {
    "BLOCK", "FUNCTION", "END_OF_STRUCT", "FILE", "SECTION", "WEAK_EXTERNAL"
};

//
                                // Given a symbol storage class value, return a descriptive ASCII string
                                //
PSTR            GetSZStorageClass(BYTE storageClass)
{
    if (storageClass <= IMAGE_SYM_CLASS_BIT_FIELD)
        return SzStorageClass1[storageClass];
    else if ((storageClass >= IMAGE_SYM_CLASS_BLOCK)
             && (storageClass <= IMAGE_SYM_CLASS_WEAK_EXTERNAL))
        return SzStorageClass2[storageClass - IMAGE_SYM_CLASS_BLOCK];
    else
        return "???";
}
int  FindFunctionStartLine(char *sname,int idx)
{
    PIMAGE_AUX_SYMBOL auxSym;
	PIMAGE_SYMBOL pSymbol;
	int result;
	char *p;

	if (idx >= NumberOfSymbols) {
		return 0;
	}
	pSymbol = &PCOFFSymbolTable[idx];
	if ((pSymbol->Type & 0xF0) != (IMAGE_SYM_DTYPE_FUNCTION << 4)) {
		return(-1);
	}
	if (pSymbol->NumberOfAuxSymbols < 1) {
		return -2;
	}
    auxSym = (PIMAGE_AUX_SYMBOL) (pSymbol + 1);

	pSymbol = &PCOFFSymbolTable[auxSym->Sym.TagIndex];
    auxSym = (PIMAGE_AUX_SYMBOL) (pSymbol + 1);
	p = (char *)auxSym;
	result = *(unsigned short *)(&p[4]);
	return(result);
}

/*
Auxiliary Symbol Table records always follow, and apply to, some standard Symbol
Table record. An auxiliary record can have any format that the tools are designed
to recognize, but 18 bytes must be allocated for them so that Symbol Table is
maintained as an array of regular size. Currently, Microsoft tools recognize
auxiliary formats for the following kinds of records: function definitions, function
begin and end symbols (.bf and .ef), weak externals, filenames, and section
definitions.
The traditional COFF design also includes auxiliary-record formats for arrays and
structures. Microsoft tools do not use these, instead placing that symbolic
information in CodeView format in the debug sections.

5.5.1. Auxiliary Format 1: Function Definitions
A symbol table record marks the beginning of a function defintion if all of the
following are true: it has storage class EXTERNAL (2), a Type value indicating it is
a function (0x20), and a section number greater than zero. Note that a symbol table
record that has a section number of UNDEFINED (0) does not define the function and
does not have an auxiliary record. Function-definition symbol records are followed
by an auxiliary record with the format described below.
Offset	Size	Field	Description
0		4		Tag Index
						Symbol-table index of the corresponding .bf (begin function)
					    symbol record.
4		4		Total Size
						Size of the executable code for the function itself. If the
						function is in its own section, the Size of Raw Data in the
						section header will be greater or equal to this field,
						depending on alignment considerations.
8		4		Pointer to Linenumbers
						File offset of the first COFF line-number entry for the
						function, or zero if none exists. See Section 5.3, 'COFF
						Line Numbers,' for more information.
12		4		Pointer to Next Function
						Symbol-table index of the record for the next function. If
						the function is the last in the symbol table, this field is
						set to zero.
16		2		Unused.

5.5.2. Auxiliary Format 2: .bf and .ef Symbols
For each function definition in the Symbol Table, there are three contiguous items
that describe the beginning, ending, and number of lines. Each of these symbols has
storage class FUNCTION (101):
n	A symbol record named .bf (begin function). The Value field is unused.
n	A symbol record named .lf (lines in function). The Value field gives the number
	of lines in the function.
n	A symbol record named .ef (end of function). The Value field has the same number
	as the Total Size field in the function-definition symbol record.
The .bf and .ef symbol records (but not .lf records) are followed by an auxiliary
	record with the following format:
Offset	Size	Field	Description
0		4		Unused.
4		2		Line Number	Actual ordinal line number (1, 2, 3, etc.) within
				source file, corresponding to the .bf or .ef record.
6		6		Unused.
12		4		Pointer to Next Function (.bf only)	Symbol-table index of the next
				.bf symbol record. If the function is the last in the symbol table,
			    this field is set to zero. Not used for .ef records.
16		2		Unused.

5.5.3. Auxiliary Format 3: Weak Externals

'Weak externals' are a mechanism for object files allowing flexibility at link time.
A module can contain an unresolved external symbol (sym1), but it can also include
an auxiliary record indicating that if sym1 is not present at link time, another
external symbol (sym2) is used to resolve references instead.
If a definition of sym1 is linked, then an external reference to the symbol is
resolved normally. If a definition of sym1 is not linked, then all references to
the weak external for sym1 refer to sym2 instead. The external symbol, sym2, must
always be linked; typically it is defined in the module containing the weak
reference to sym1.

Weak externals are represented by a Symbol Table record with EXTERNAL storage class,
UNDEF section number, and a value of 0. The weak-external symbol record is followed
by an auxiliary record with the following format:
Offset	Size	Field	Description
0		4		Tag Index
						Symbol-table index of sym2, the symbol to be linked if sym1
					    is not found.
4		4		Characteristics
						A value of 1 indicating that no library search for sym1
						should be performed, or a value of 2 indicating that the
						linker should search all libraries.

8		10		Unused.
Note that the Characteristics field is not defined in WINNT.H; instead, the Total
Size field is used.

5.5.4. Auxiliary Format 4: Files
This format follows a symbol-table record with storage class FILE (103). The symbol
name itself should be .file, and the auxiliary record that follows it gives the name
of a source-code file.
Offset	Size	Field	Description
0		18		File Name	ASCII string giving the name of the source file;
						padded with nulls if less than maximum length.

5.5.5. Auxiliary Format 5: Section Definitions
This format follows a symbol-table record that defines a section: such a record has
a symbol name that is the name of a section (such as .text or .drectve) and has
storage class STATIC (3). The auxiliary record provides information on the section
referred to. Thus it duplicates some of the information in the section header.
Offset	Size	Field	Description
0		4		Length	Size of section data; same as Size of Raw Data in the
						section header.
4		2		Number Of Relocs
						Number of relocation entries for the section.
6		2		Number Of Linenumbers
						Number of line-number entries for the section.
8		4		Check Sum
						Checksum for communal data. Applicable if the
						IMAGE_SCN_LNK_COMDAT flag is set in the section header. See
					    'COMDAT Sections' below, for more information.
12		2		Number	One-based index into the Section Table for the associated
						section; used when the COMDAT Selection setting is 5.
14		1		Selection
						COMDAT selection number. Applicable if the section is a
						COMDAT section.
15		3		Unused.

5.5.6. COMDAT Sections (Object Only)
The Selection field of the Section Definition auxiliary format is applicable if the
section is a COMDAT section: a section that can be defined by more than one object
file. (The flag IMAGE_SCN_LNK_COMDAT is set in the Section Flags field of the
section header.) The way that the linker resolves the multiple definitions of
COMDAT sections is determined by the Selection field.
The first symbol having the section value of the COMDAT section is the section
symbol This symbol has the name of the section, Value field equal to 0, the section
number of the COMDAT section in question, Type field equal to IMAGE_SYM_TYPE_NULL,
Class field equal to IMAGE_SYM_CLASS_STATIC, and one auxiliary record. The second
symbol is called 'the COMDAT symbol' and is used by the linker in conjunction with
the Selection field.
Values for the Selection field are shown below.
 Constant		Value	Description

IMAGE_COMDAT_SELECT_NODUPLICATES
				1		The linker generates a warning if more than one section
						defines the same COMDAT symbol, but links in one of the
						sections anyway.
IMAGE_COMDAT_SELECT_ANY
				2		Any section defining the same COMDAT symbol may be linked;
						the rest are removed.
IMAGE_COMDAT_SELECT_SAME_SIZE
				3		The linker chooses an arbitrary section among the duplicate
						sections (having same COMDAT symbol); however, all must be
						the same size or the linker generates a warning.
IMAGE_COMDAT_SELECT_EXACT_MATCH
				4		All duplicate sections must match exactly. One of them is
						linked. (This is not currently implemented in the linker).
IMAGE_COMDAT_SELECT_ASSOCIATIVE
				5		The section is linked if a certain other COMDAT section is
						linked. This other section is indicated by the Number field
						of the auxiliary symbol record for the section definition.
						Use of this setting is useful for definitions that have
						components in multiple sections (for example, code in one
						and data in another), but where all must be linked together.
*/
int  DumpAuxSymbols(PIMAGE_SYMBOL pSymbolTable,char *sname,int nb)
{
    PIMAGE_AUX_SYMBOL auxSym;
	int isFunction = 0;
	char *p;
	char tmpbuf[10*sizeof(IMAGE_SYMBOL)+1];

    auxSym = (PIMAGE_AUX_SYMBOL) (pSymbolTable + 1);

    if (pSymbolTable->StorageClass == IMAGE_SYM_CLASS_FILE) {
		memset(tmpbuf,0,sizeof(tmpbuf));
		memcpy(tmpbuf,auxSym,nb*sizeof(IMAGE_SYMBOL));
        printf("     * %s\n", tmpbuf);
		strcpy(SourceFileName,(char *)tmpbuf);
	}
    else if ((pSymbolTable->StorageClass == IMAGE_SYM_CLASS_EXTERNAL)) {
        if ((pSymbolTable->Type & 0xF0) == (IMAGE_SYM_DTYPE_FUNCTION << 4)) {
isfn:
            printf("     * tag: %05ld  size: %05ld  Line #'s: %09ld  next fn: %05ld\n",
                   auxSym->Sym.TagIndex, auxSym->Sym.Misc.TotalSize,
                   auxSym->Sym.FcnAry.Function.PointerToLinenumber,
                   auxSym->Sym.FcnAry.Function.PointerToNextFunction);
			isFunction = 1;
        }
    }
    else if ((pSymbolTable->StorageClass == IMAGE_SYM_CLASS_STATIC)) {
		if (pSymbolTable->Type == 32) goto isfn;
        printf(
          "     * Section: %04d  Len: %05ld  Relocs: %05d  LineNums: %05d\n",
               auxSym->Section.Number, auxSym->Section.Length,
               auxSym->Section.NumberOfRelocations,
               auxSym->Section.NumberOfLinenumbers);
    }
	else if (!strncmp(sname,".bf",3)) {
		p = (char *)auxSym;
		printf("    * Fn begins line %d. Next fn:%ld\n",
				*(unsigned short *)(&p[4]),*(unsigned long *)(&p[12]));
	}
	else if (!strncmp(sname,".ef",3)) {
		p = (char *)auxSym;
		printf("    * %4ld %4ld %4ld %4d Function end line #%d\n",
			*(unsigned long *)(p),
			*(unsigned long *)(&p[8]),
			*(unsigned long *)(&p[12]),
			*(unsigned short *)(&p[16]),
			*(unsigned short *)(&p[4]));
	}
	return(isFunction);
}

// Given a COFF symbol table index, look up its name.  This function assumes
// that the COFFSymbolCount and PCOFFSymbolTable variables have been already
// set up.
//
BOOL            LookupSymbolName(DWORD index, PSTR buffer, UINT length)
{
    PSTR            stringTable;

    if (index >= COFFSymbolCount)
        return FALSE;

    if (PCOFFSymbolTable == 0)
        return FALSE;

	if (index > (DWORD)NumberOfSymbols) {
		sprintf(buffer,"Illegal index %ld",index);
		return(FALSE);
	}
    if (PCOFFSymbolTable[index].N.Name.Short != 0) {
        strncpy(buffer, PCOFFSymbolTable[index].N.ShortName, min(8, length));
        buffer[8] = 0;
    }
    else {
        stringTable = (PSTR) & PCOFFSymbolTable[COFFSymbolCount];
        strncpy(buffer,
                stringTable + PCOFFSymbolTable[index].N.Name.Long, length);
        buffer[length - 1] = 0;
    }

    return TRUE;
}
/*
Dumps the string table
----------------------
Immediately following the COFF symbol table is the COFF string table. The position
of this table is found by taking the symbol table address in the COFF header, and
adding the number of symbols multiplied by the size of a symbol.
At the beginning of the COFF string table are 4 bytes containing the total size (in
bytes) of the rest of the string table.  This size does not include the Size field
itself, so that the value in this location would be 0 if no strings were present.
Following the size are null-terminated strings pointed to by symbols in the COFF
symbol table.
*/
                                //
void            DumpStringTable(char *pstringTable)
{
    char           *stop;
    unsigned long   len;
    int             i, count, s;

    stringTable = pstringTable;
	fflush(stdout);
    printf("\nString Table\n------------\n");
    len = *(unsigned long *) pstringTable;
	if (len > (unsigned long)TotalFileSize) {
		printf("********Incorrect length field of String table: %ld\n",len);
		fflush(stdout);
		exit(1);
	}
    printf("Length: %ld(d)\n", len);
    stop = pstringTable + len;
    count = 4;
    pstringTable += 4;
    i = 1;
    while (pstringTable < stop) {
        s = strlen(pstringTable) + 1;
        printf("[%3d %5d %5d] \"%s\"\n", i, count, count + s, pstringTable);
        i++;
        pstringTable += s;
        count += s;
    }
}

/*
Dumps a COFF symbol table from an EXE or OBJ

The Symbol Table described in this section is inherited from the traditional COFF
format. It is distinct from CodeView information. A file may contain both a COFF
Symbol Table and CodeView debug information, and the two are kept separate. Some
Microsoft tools use the Symbol Table for limited but important purposes, such as
communicating COMDAT information to the linker. Section names and file names, as
well as code and data symbols, are listed in the Symbol Table.
The location of the Symbol Table is indicated in the COFF Header.
The Symbol Table is an array of records, each 18 bytes long. Each record is either a
standard or auxiliary symbol-table record. A standard record defines a symbol or
name, and has the following format:
Offset	Size	Field		Description

0		8		Name (*)	Name of the symbol, represented by union of three
							structures. An array of eight bytes is used if the name
							is not more than eight bytes long. See Section 5.4.1,
							'Symbol Name Representation,' for more information.
8		4		Value		Value associated with the symbol. The interpretation of
							this field depends on Section Number and Storage Class.
						    A typical meaning is the relocatable address.
12		2		Section Number
							Signed integer identifying the section, using a
							one-based index into the Section Table. Some values have
						    special meaning defined in 'Section Number Values.'
14		2		Type		A number representing type. Microsoft tools set this
							field to 0x20 (function) or 0x0 (not a function). See
							Section 5.4.3, 'Type Representation,' for more
							information.
16		1		Storage Class	Enumerated value representing storage class. See
							Section 5.4.4, 'Storage Class,' for more information.
17		1		Number of Aux Symbols
							Number of auxiliary symbol table entries that follow
							this record.
Zero or more auxiliary symbol-table records immediately follow each standard
symbol-table record. However, typically not more than one auxiliary symbol-table
record follows a standard symbol-table record (except for .file records with long
file names). Each auxiliary record is the same size as a standard symbol-table
record (18 bytes), but rather than define a new symbol, the auxiliary record gives
additional information on the last symbol defined. The choice of which of several
formats to use depends on the Storage Class field. Currently-defined formats for
auxiliary symbol table records are shown in 'Auxiliary Symbol Records'

Tools that read COFF symbol tables must ignore auxiliary symbol records whose
interpretation is unknown. This allows the symbol table format to be extended to
add new auxiliary records, without breaking existing tools.

5.4.1. Symbol Name Representation
The Name field in a symbol table consists of eight bytes that contain the name itself,
if not too long, or else give an offset into the String Table. To determine whether
the name itself or an offset is given, test the first four bytes for equality to
zero.
Offset	Size	Field	Description

0		8		Short Name
						An array of eight bytes. This array is padded with nulls on
						the right if the name is less than eight bytes long.
0		4		Zeroes	Set to all zeros if the name is longer than eight bytes.
4		4		Offset	Offset into the String Table.

5.4.2. Section Number Values
Normally, the Section Value field in a symbol table entry is a one-based index into
the Section Table. However, this field is a signed integer and may take negative
values. The following values, less than one, have special meanings:

Constant			Value	Description

IMAGE_SYM_UNDEFINED	0		Symbol record is not yet assigned a section. In object
							files, this setting is used for  references to external
							symbols not defined by the symbol record.
IMAGE_SYM_ABSOLUTE	-1		The symbol has a value but is not an address. Examples
							are automatic variables and registers. Not used by all
							Microsoft tools.
IMAGE_SYM_DEBUG		-2		The symbol provides general type or debugging
							information but does not correspond to a section.
							Microsoft tools use this setting along with .file
							records (storage class FILE).
5.4.3. Type Representation
The Type field of a symbol table entry contains two bytes, each byte representing type information. The least-significant byte represents simple (base) data type, and the most-significant byte represents complex type, if any:
MSB	LSB

Complex type: none, pointer, function, array.	Base type: integer, floating-point,
etc.
The following values are defined for base type, although Microsoft tools generally
do not use this field, setting the least-significant byte to 0. Instead, CodeView
information is used to indicate types. However, the possible COFF values are listed
here for completeness.

Constant	Value	Description

IMAGE_SYM_TYPE_NULL	0	No type information, or unknown base type. Microsoft tools
						use this setting.
IMAGE_SYM_TYPE_VOID	1	No valid type; used with void pointers and functions.
IMAGE_SYM_TYPE_CHAR	2	Character (signed byte).
IMAGE_SYM_TYPE_SHORT	3	Two-byte signed integer.
IMAGE_SYM_TYPE_INT	4	Natural integer type (normally four bytes in Windows NT).
IMAGE_SYM_TYPE_LONG	5	Four-byte signed integer.
IMAGE_SYM_TYPE_FLOAT	6	Four-byte floating-point number.
IMAGE_SYM_TYPE_DOUBLE	7	Eight-byte floating-point number.
IMAGE_SYM_TYPE_STRUCT	8	Structure.
IMAGE_SYM_TYPE_UNION	9	Union.
IMAGE_SYM_TYPE_ENUM	10	Enumerated type.
IMAGE_SYM_TYPE_MOE	11	Member of enumeration (a specific value).
IMAGE_SYM_TYPE_BYTE	12	Byte; unsigned one-byte integer.
IMAGE_SYM_TYPE_WORD	13	Word; unsigned two-byte integer.
IMAGE_SYM_TYPE_UINT	14	Unsigned integer of natural size (normally, four bytes).
IMAGE_SYM_TYPE_DWORD	15	Unsigned four-byte integer.

The  most significant byte specifies whether the symbol is a pointer to, function
returning, or array of the base type specified in the least significant byte.
Microsoft tools use this field only to indicate whether or not the symbol is a
function, so that the only two resulting values are 0x0 and 0x20 for the Type field.
However, other tools can use this field to communicate more information.
Constant	Value	Description

IMAGE_SYM_DTYPE_NULL	0	No derived type; the symbol is a simple scalar variable.
IMAGE_SYM_DTYPE_POINTER	1	Pointer to base type.
IMAGE_SYM_DTYPE_FUNCTION	2	Function returning base type.
IMAGE_SYM_DTYPE_ARRAY	3	Array of base type.

5.4.4. Storage Class
The Storage Class field of the Symbol Table indicates what kind of definition a
symbol represents. The following table shows possible values. Note that the Storage
Class field is an unsigned one-byte integer. The special value -1 should therefore
be taken to mean its unsigned equivalent, 0xFF.
Although traditional COFF format makes use of many storage-class values, Microsoft
tools rely on CodeView format for most symbolic information and generally use only
four storage-class values: EXTERNAL (2), STATIC (3), FUNCTION (101), and STATIC (103).
Except in the second column heading below, 'Value' should be taken to mean the
Value field of the symbol record (whose interpretation depends on the number found
as the storage class).
Constant						Value	Description / Interpretation of Value Field

IMAGE_SYM_CLASS_END_OF_FUNCTION	-1 (0xFF)
										Special symbol representing end of function,
									    for debugging purposes.
IMAGE_SYM_CLASS_NULL			0		No storage class assigned.
IMAGE_SYM_CLASS_AUTOMATIC		1		Automatic (stack) variable. The Value field
										specifies stack frame offset.
IMAGE_SYM_CLASS_EXTERNAL		2		Used by Microsoft tools for external
										symbols. The Value field indicates the size
										if the section number is IMAGE_SYM_UNDEFINED
									    (0). If the section number is not 0, then
										the Value field specifies the offset within
										the section.
IMAGE_SYM_CLASS_STATIC			3		The Value field specifies the offset of the
										symbol within the section. If the Value is 0
										then the symbol represents a section name.
IMAGE_SYM_CLASS_REGISTER		4		Register variable. The Value field specifies
										register number.
IMAGE_SYM_CLASS_EXTERNAL_DEF	5		Symbol is defined externally.
IMAGE_SYM_CLASS_LABEL			6		Code label defined within the module. The
										Value field specifies the offset of the
										symbol within the section.
IMAGE_SYM_CLASS_UNDEFINED_LABEL	7		Reference to a code label not defined.
IMAGE_SYM_CLASS_MEMBER_OF_STRUCT	8	Structure member. The Value field specifies
										nth member.
IMAGE_SYM_CLASS_ARGUMENT		9		Formal argument (parameter)of a function.
										The Value field specifies nth argument.
IMAGE_SYM_CLASS_STRUCT_TAG		10		Structure tag-name entry.
IMAGE_SYM_CLASS_MEMBER_OF_UNION	11		Union member. The Value field specifies nth
										member.
IMAGE_SYM_CLASS_UNION_TAG		12		Union tag-name entry.
IMAGE_SYM_CLASS_TYPE_DEFINITION	13		Typedef entry.
IMAGE_SYM_CLASS_UNDEFINED_STATIC	14	Static data declaration.
IMAGE_SYM_CLASS_ENUM_TAG		15		Enumerated type tagname entry.
IMAGE_SYM_CLASS_MEMBER_OF_ENUM	16		Member of enumeration. Value specifies nth
										member.
IMAGE_SYM_CLASS_REGISTER_PARAM	17		Register parameter.
IMAGE_SYM_CLASS_BIT_FIELD		18		Bit-field reference. Value specifies nth bit
										in the bit field.
IMAGE_SYM_CLASS_BLOCK			100		A .bb (beginning of block) or .eb (end of
										block) record. Value is the relocatable
										address of the code location.
IMAGE_SYM_CLASS_FUNCTION		101		Used by Microsoft tools for symbol records
										that define the extent of a function: begin
										function (named .bf), end function (.ef), and
									    lines in function (.lf). For .lf records,
										Value gives the number of source lines in the
									    function. For .ef records, Value gives the
										size of function code.
IMAGE_SYM_CLASS_END_OF_STRUCT	102		End of structure entry.
IMAGE_SYM_CLASS_FILE			103		Used by Microsoft tools, as well as
										traditional COFF format, for the source-file
									    symbol record. The symbol is followed by an
										auxiliary record that names the file.
IMAGE_SYM_CLASS_SECTION			104		Definition of a section (Microsoft tools use
										STATIC storage class instead).
IMAGE_SYM_CLASS_WEAK_EXTERNAL	105		Weak external. Microsoft tools use EXTERNAL
										storage class instead. See Section 5.5.3,
										'Auxiliary Format 3: Weak Externals,' for
										more information.
*/
int NumberOfFunctions = 0;                                //
PIMAGE_SYMBOL *FunctionIndexTable;
#define MAXFUNCTIONS 5000
void DumpSymbolTable(PIMAGE_SYMBOL p, unsigned cSymbols)
{
    unsigned        i,val;
    PSTR            stringTable;
    char            sectionName[10],*name;

	if (fShowSymbolTable)
    printf("Symbol Table - %d entries of %d bytes each (* = auxiliary symbol)\n", cSymbols,sizeof(IMAGE_SYMBOL));
    pSymbolTable = p;
	FunctionIndexTable = (PIMAGE_SYMBOL *)malloc(MAXFUNCTIONS*sizeof(int));
	if (fShowSymbolTable)
    printf(
     "Indx Name                 Value    Section    cAux  Type    Storage\n"
           "---- -------------------- -------- ---------- ----- ------- --------\n");

    // The string table apparently starts right after the symbol table
    stringTable = (PSTR) & pSymbolTable[cSymbols];
	if (!IsGoodPointer(stringTable)) {
		printf("\n***Incorrect number of symbols in module %d\n",cSymbols);
		exit(1);
	}
    if ((*(unsigned long *)stringTable) > (unsigned long)TotalFileSize) {
	printf("\n***Incorrect length of string table: %ld\n",(*(unsigned long *)stringTable));
	exit(1);
    }
    for (i = 0; i < cSymbols; i++) {
	if (fShowSymbolTable)
        printf("%04d ", i);

        if (pSymbolTable->N.Name.Short != 0) {
		if (fShowSymbolTable)
            printf("%-20.8s", pSymbolTable->N.ShortName);
			name = pSymbolTable->N.ShortName;
		}
        else {
			name = stringTable + pSymbolTable->N.Name.Long;
            if (strlen(name) < 21) {
		if (fShowSymbolTable)
            	printf("%-20s", name);
		}
            else {
		if (fShowSymbolTable) {
                printf("%s\n",name);
                printf("%25s"," ");
		}
		}
		}
		val = pSymbolTable->Value;
		/* reverse bytes if necessary */
		if (pImageFileHeader->Characteristics & 0x80) {
			val = (val & 0xFFFF0000) | (0xFF & (val >> 8)) | (0xFF00 & (val << 8));
		}
		if (pImageFileHeader->Characteristics & 0x8000) {
			val = (val & 0xFFFF) | (0xFF0000 & (val >> 8)) | (0xFF000000 & (val << 8));
		}
	if (fShowSymbolTable)
        printf(" %08d", val);

        GetSectionName(pSymbolTable->SectionNumber, sectionName,
                       sizeof(sectionName));
	if (fShowSymbolTable)
        printf(" sect:%s aux:%d type:%02d st:%s\n",
               sectionName,
               pSymbolTable->NumberOfAuxSymbols,
               pSymbolTable->Type,
               GetSZStorageClass(pSymbolTable->StorageClass));

        if (pSymbolTable->NumberOfAuxSymbols && fShowSymbolTable) {
            if (DumpAuxSymbols(pSymbolTable,name,pSymbolTable->NumberOfAuxSymbols)) {
				FunctionIndexTable[NumberOfFunctions] = pSymbolTable;
				NumberOfFunctions++;
			}
		}

        // Take into account any aux symbols
        i += pSymbolTable->NumberOfAuxSymbols;
        pSymbolTable += pSymbolTable->NumberOfAuxSymbols;
        pSymbolTable++;
    }
    if (fShowSymbolTable) DumpStringTable(stringTable);
    pSymbolTable = p;
}

//
                                // Given a section name, look it up in the section table and return a
                                // pointer to the start of its raw data area.
                                //
LPVOID GetSectionPtr(PSTR name, PIMAGE_NT_HEADERS pNTHeader, DWORD imageBase)
{
    PIMAGE_SECTION_HEADER section;
    unsigned        i;

    section = (PIMAGE_SECTION_HEADER) (pNTHeader + 1);

    for (i = 0; i < pNTHeader->FileHeader.NumberOfSections; i++, section++) {
        if (_strnicmp(section->Name, name, IMAGE_SIZEOF_SHORT_NAME) == 0)
            return (LPVOID) (section->PointerToRawData + imageBase);
    }

    return 0;
}

//
// Given a section name, look it up in the section table and return a
// pointer to its IMAGE_SECTION_HEADER
//
PIMAGE_SECTION_HEADER GetSectionHeader(PSTR name, PIMAGE_NT_HEADERS pNTHeader)
{
    PIMAGE_SECTION_HEADER section;
    unsigned        i;

    section = (PIMAGE_SECTION_HEADER) (pNTHeader + 1);

    for (i = 0; i < pNTHeader->FileHeader.NumberOfSections; i++, section++) {
        if (_strnicmp(section->Name, name, IMAGE_SIZEOF_SHORT_NAME) == 0)
            return section;
    }

    return 0;
}

/*
 Do a hexadecimal dump of the raw data for all the sections.  You
 could just dump one section by adjusting the PIMAGE_SECTION_HEADER
 and cSections parameters
*/
void            DumpRawSectionData(PIMAGE_SECTION_HEADER section,
                                   PVOID base,
                                   unsigned cSections)
{
  unsigned        i;
  char            name[IMAGE_SIZEOF_SHORT_NAME + 1];
  
  printf("Section Hex Dumps\n");
  
  for (i = 0; i < cSections; i++, section++) {
    /* Make a copy of the section name so that we can ensure that */
    /* it's null-terminated */
    memcpy(name, section->Name, IMAGE_SIZEOF_SHORT_NAME);
    name[IMAGE_SIZEOF_SHORT_NAME] = 0;
    
    /* Don't dump sections that don't exist in the file! */
    if (section->PointerToRawData == 0)
      continue;
    
    printf("section %02d (%s)  size: %05ld  file offs: %05ld\n",
	   i, name,
	   section->SizeOfRawData,
	   section->PointerToRawData);
    if (!strcmp(name, ".text") && fShowRawSectionData) {
      DoDisassemble(0, section->SizeOfRawData, MakePtr(PBYTE, base, section->PointerToRawData));
      printf("\nNumber of Functions: %d\n",NumberOfFunctions);
      
    }
    else	if (!strncmp(name,".debug$S",8)) {
      if (fDebug)
	DumpCodeViewSymbols(MakePtr(PBYTE, base, section->PointerToRawData),
			    section->SizeOfRawData,NULL,1);
    }
    else	if (!strncmp(name,".debug$T",8)) {
      if (fDebug)
	DumpCodeViewTypes(MakePtr(PBYTE, base, section->PointerToRawData),
			  section->SizeOfRawData,1,NULL);
    }
    else if (!strncmp(name,".rsrc$01",8)) {
      printf("Resources\n");
      pResourceData = NULL;
      DumpResourceDirectory(
			    (PIMAGE_RESOURCE_DIRECTORY)MakePtr(PBYTE,
							       base,section->PointerToRawData),
			    (DWORD)MakePtr(PBYTE,base,section->PointerToRawData),0,0);
      goto hexa;
    }
    else
hexa:
    if (fShowRawSectionData)
      HexDump(MakePtr(PBYTE, base, section->PointerToRawData),
	      section->SizeOfRawData);
    
    printf("\n");
  }
}

// Dump a range of line numbers from the COFF debug information
typedef struct tagLinNrOffset {
  int LnNumber;
  int Offset;
} COFF_LINENR;

COFF_LINENR *LineNumberTable;
/*
COFF line numbers indicate the relationship between code and line-numbers in source
files. The Microsoft format for COFF line numbers is similar to standard COFF, but
it has been extended to allow a single section to relate to line numbers in multiple
source files.
COFF line numbers consist of an array of fixed-length records. The location (file
offset) and size of the array are specified in the section header. Each line-number
record is of the following format:
Offset	Size	Field		Description

0		4		Type (*)	Union of two fields: Symbol Table Index and RVA. Whether
							Symbol Table Index or RVA is used depends on the value
							of Linenumber.
4		2		Linenumber	When nonzero, this field specifies a one-based line
							number. When zero, the Type field is interpreted as a
							Symbol Table Index for a function.
The Type field is a union of two four-byte fields, Symbol Table Index, and RVA:

Offset	Size	Field	Description
0		4		Symbol Table Index	Used when Linenumber is 0: index to symbol
				table entry for a function. This format is used to indicate the
				function that a group of line-number records refer to.
0		4		RVA	Used when Linenumber is greater than 0: relative virtual address
				of the executable code that corresponds to the source line indicated
				(In an object file, the RVA is relative to the section.)

A line-number record, then, can either set the Linenumber field to 0 and point to a
function definition in the Symbol Table, or else it can work as a standard
line-number entry by giving a positive integer (line number) and the corresponding
address in the object code.

A group of line-number entries always begins with the first format: the index of a
function symbol. If this is the first line-number record in the section, then it is
also the COMDAT symbol name for the function if the sections COMDAT flag is set.
(See Section 5.5.6, 'COMDAT Sections.') The functions auxiliary record in the
Symbol Table has a Pointer to Linenumbers field that points to this same
line-number record.

A record identifying a function is followed by any number of line-number entries
that give actual line-number information (Linenumber greater than zero). These
entries are one-based, relative to the beginning of the function, and represent
every source line in the function except for the first one.

For example, the first line-number record for the following example would specify
the ReverseSign function (Symbol Table Index of ReverseSign, Linenumber set to 0).
Then records with Linenumber values of 1, 2, and 3 would follow, corresponding to
source lines as shown:

    // some code precedes ReverseSign function

	int	ReverseSign(int i)
1:	{
2:		return -1 * i;
3:	}
*/
void DumpLineNumbers(PIMAGE_LINENUMBER pln, DWORD count, DWORD offset)
{
    char            buffer[64];
    DWORD           i,StartLineNumber=0;
    int l;
    
    if (count == 0) return;
    if (fShowLineNumbers)
      printf("Line Numbers at offset %ld in file.\n",offset);
    if (offset == 0) return;
    if (LineNumberTable == NULL) {
      LineNumberTable = (COFF_LINENR *)malloc(pCodeSection->NumberOfLinenumbers * sizeof(COFF_LINENR));
      ObjLineNumberIdx = 0;
    }
    for (i = 0; i < count; i++) {
      if (pln->Linenumber == 0) { /* A symbol table index */
	
	buffer[0] = 0;
	if (!LookupSymbolName(pln->Type.SymbolTableIndex, buffer,
			      sizeof(buffer))) {
	  printf("Incorrect symbol index %ld in line numbers!\n",
		 pln->Type.SymbolTableIndex);
	  break;
	}
      if (fShowLineNumbers)
	printf("SymIndex: %ld (%s): offset in file %ld",
	       pln->Type.SymbolTableIndex,
	       buffer,offset + (i*sizeof(*pln)));
      StartLineNumber = FindFunctionStartLine(buffer,pln->Type.SymbolTableIndex);
	if ((int)StartLineNumber < 0) {
	  printf(" Illegal index in the symbol table");
	}
	if (fShowLineNumbers)
	  printf("\n");
    }
    else{                    /* A regular line number */
				  l = StartLineNumber+pln->Linenumber;
    
		if (fShowLineNumbers)
            printf(" Addr: %6ld  Line: [%4d %6d]\n",
                   pln->Type.VirtualAddress, pln->Linenumber,l );
			LineNumberTable[ObjLineNumberIdx].LnNumber = l;
			LineNumberTable[ObjLineNumberIdx].Offset = pln->Type.VirtualAddress;
			ObjLineNumberIdx++;
		}
        pln++;
		offset += 6;
    }
	printf("Line numbers end at file position %ld\n",offset);
}

// Number of hex values displayed per line
#define HEX_DUMP_WIDTH 16

//
                                // Dump a region of memory in a hexadecimal format
                                //
void            HexDump(PBYTE ptr, DWORD length)
{
    char            buffer[256];
    PSTR            buffPtr;
    unsigned        cOutput, i;
    DWORD           bytesToGo = length;

    while (bytesToGo) {
        cOutput = bytesToGo >= HEX_DUMP_WIDTH ? HEX_DUMP_WIDTH : bytesToGo;

        buffPtr = buffer;
        buffPtr += sprintf(buffPtr, "%08ld:  ", length - bytesToGo);

        for (i = 0; i < HEX_DUMP_WIDTH; i++) {
            buffPtr += sprintf(buffPtr,
                                i < cOutput ? "%02X " : "   ",
                                *(ptr + i));

            // Put an extra space between the 1st and 2nd half of the bytes
            // on each line.
            if (i == (HEX_DUMP_WIDTH / 2) - 1)
                buffPtr += sprintf(buffPtr, " ");
        }

        for (i = 0; i < cOutput; i++) {
            char            c;

            c = '.';
            if (isascii(*(ptr + i)))
                c = isprint(*(ptr + i)) ? *(ptr + i) : '.';

            buffPtr += sprintf(buffPtr, "%c", c);
        }
        puts(buffer);           // Can't use printf(), since there may be a '%'
        // in the string.

        bytesToGo -= cOutput;
        ptr += HEX_DUMP_WIDTH;
    }
}
static char *indent = "      ";


void DisplayLengthPrefixedString(char *p)
{
	int l = *p++;

	while (l > 0) {
		printf("%c",*p++);
		l--;
	}
}
void DisplayObjName(char *data)
{
	printf("Obj signature %lX, name '",*(unsigned long *)data);
	DisplayLengthPrefixedString(data+4);
	printf("'\n");
}

static void DisplayBlockStart(char *data)
{
	struct blockInfo *s;

	s = (struct blockInfo *)data;
	printf("Lexical block ");
	DisplayLengthPrefixedString(s->name);
	printf("\n%sParent %ld End %ld length %ld offset %ld segment %d\n",indent,
			s->pParent,s->pEnd,s->length,s->offset,s->segment);
}

static void DisplayUDT(char *data)
{
	struct UDT *s;

	s = (struct UDT *)data;
	printf("User defined type '");
	DisplayLengthPrefixedString(s->name);
	printf("'\n%sType %d\n",indent,s->type);
}
static void DisplayCodeLabel(char *data)
{
	struct codeLabel *s;

	s = (struct codeLabel *)data;
	printf("Code Label ");
	DisplayLengthPrefixedString(s->name);
	printf("\n%sOffset %ld Segment %d Flags 0x%X\n",indent,s->offset,s->segment,s->flags);
}

static struct procInfo *DisplayProcedureStart(int idx,char *data)
{
	struct procInfo *s,*ns;
	char *n;

	s = (struct procInfo *)data;
	ns = (struct procInfo *)malloc(sizeof(struct procInfo));
	memcpy(ns,data,sizeof(struct procInfo));
	if (idx == S_LPROC32) {
		printf("Local ");
	}
	else {
		printf("Global ");
	}
	printf("Procedure '");
	DisplayLengthPrefixedString(s->u.fname);
	n = malloc(s->u.fname[0]+1);
	memset(n,0,s->u.fname[0]+1);
	memcpy(n,s->u.fname+1,s->u.fname[0]);
	ns->u.Name = n;
	printf("'\n%sParent %ld End %ld Next %ld Length %ld debStart %ld debEnd %ld offset %ld\n",
		indent,s->pParent,s->pEnd,s->pNext,s->procLength,s->debugStart,s->debugEnd,s->offset);
	printf("%sSegment %d proc record type index %d(%d) flags 0x%XH\n",indent,
		s->segment,s->procType,s->procType-4096,s->flags);
	return(ns);
}

static void DisplayVariable(int idx,char *data)
{
	struct localDataInfo *s;

	s = (struct localDataInfo *)data;
	if (idx == S_LDATA32)
		printf("Static '");
	else if (idx == S_GDATA32) printf("Global '");
	else if (idx == S_PUB32) printf("Public symbol ");
	else {
		printf("Unknown variable type %d ",idx);
	}
	DisplayLengthPrefixedString(s->name);
	printf("'\n%sOffset %ld segment %d type %d\n",indent,s->offset,s->segment,s->type);
}

static void DisplayAutoVariable(char *data)
{
	struct autoDataInfo *s;

	s = (struct autoDataInfo *)data;
	printf("Automatic '");
	DisplayLengthPrefixedString(s->name);
	printf("'\n%sEBP offset %ld type: '%s'\n",indent,s->offset,DisplayType(s->type));
}

void DisplayCompilerInfo(char *ptr)
{
	printf("Compile Info:\n%sMachine 0x%X Language %d Flags1 0x%X Flags2 0x%X\n%s'",
		indent,*ptr,ptr[1],ptr[2],ptr[3],indent);
	DisplayLengthPrefixedString(ptr+4);
	printf("'\n");

}

static char *regNames[] = {
	"eax","ecx","edx","ebx","esp","ebp","esi","edi"};
void DisplayRegister(char *ptr)
{
	int regNum;

	regNum = *(unsigned short *)(ptr+2);
	printf("Register variable '");
	DisplayLengthPrefixedString(ptr+4);
	printf("'\n%s type: %d, register number %d",indent,*(unsigned short *)ptr,
		regNum);
	if (regNum < 17 || regNum > 24)
		printf("***BAD***");
	else printf(" (%s)",regNames[regNum-17]);
	printf("\n");
}

void DisplayStartOfSearch(char *data)
{
	printf("Start Search: First symbol offset %ld, segment %d\n",
			*(unsigned long *)(data),*(unsigned short *)(&data[4]));
}

void DisplayCVPack(char *data,int typ)
{
	if (typ == S_PROCREF) printf("Procedure ref");
	else printf("Data ref");
	printf("(cvpack)\n%schecksum %ld offset %ld module %d\n",indent,*(unsigned long *)data,
			*(unsigned long *)(&data[4]),*(unsigned short *)(&data[8]));
}
void DisplayThunk32(char *data)
{
	unsigned long parent,end,next,offset;
	unsigned short segment,len;
	char nn;

	parent = *(unsigned long *)data;
	data += 4;
	end = *(unsigned long *)data;
	data += 4;
	next = *(unsigned long *)data;
	data += 4;
	offset = *(unsigned long *)data;
	data += 4;
	segment = *(unsigned short *)data;
	data += 2;
	len = *(unsigned short *)data;
	data += 2;
	nn = *data++;
	printf("Thunk32:\n%sparent %ld end %ld next %ld ",indent,parent,end,next);
	printf("offset %ld segment %d len %d data %d\n%sName: ",offset,segment,len,nn,indent);
	DisplayLengthPrefixedString(data);
	printf("\n");
}
/*
2.1. General
Format of Symbol Records
Data in the $$SYMBOLS segment is a stream of variable length records with the
general format:
2		2		*
length	index	data...

length	Length of record, excluding the length field
index	Type of symbol
data	Data specific to each symbol format
The symbol records are described below.  Numbers above the fields indicate the
length in bytes, and * means variable length for that field.
Symbol indices are broken into five ranges.  The first range is for symbols whose
format does not change with the compilation model of the program or the target
machine.  These include register symbols, user-defined type symbols, and so on.
The second range of symbols are those that contain 16:16 segmented addresses.  The
third symbol range is for symbols that contain 16:32 addresses.  Note that for flat
model programs, the segment is replaced with the section number for PE format exe
files..  The fourth symbol range is for symbols which are specific to the MIPS
architecture/compiler. The fifth range is for CodeView optimization.
The symbol records are formatted such that most fields fall into natural alignment
if the symbol length field is placed on a long word boundary.   For all symbols, the
variable length data is at the end of the symbol structure.  Note specifically that
fields that contain data in potentially nonaligned numeric fields must either pay
the load penalty or first do a byte wise copy of the data to a memory that is in
natural alignment.
16:16 compilers do not have to emit padding bytes between symbols to maintain natural
alignment.  The CVPACK utility places the symbols into the executable files in
natural alignment and zero pads the symbol to force alignment.  The length of each
symbol is adjusted to account for the pad bytes.  16:32 compilers must align symbols
on a long word boundary.
Provisions for enabling future implementation of register tracking and a stack
machine to perform computation on symbol addresses are provided in the symbols.
When the symbol processor is examining a symbol, the length field of the symbol is
compared with the offset of the byte following the end of the symbol name field.

If these are the same, there is no stack machine code at the end of the symbol.
If the length and offset are different, the byte following the end of the symbol
name is examined.  If the byte is zero, there is no stack machine code following the
symbol.  If the byte is not zero, then the byte indexes into the list of stack
machine implementations and styles of register tracking.  If stack machine code is
present, the address field of the symbol becomes the initial value of the stack
machine.  Microsoft does not currently emit or process stack machine code or register t
racking information.  The opcodes and operation of the stack machine has not been
defined.
Symbol Indices

0x0001	S_COMPILE	Compile flags symbol
0x0002	S_REGISTER	Register variable
0x0003	S_CONSTANT	Constant symbol
0x0004	S_UDT	User-defined Type
0x0005	S_SSEARCH	Start search
0x0006	S_END	End block, procedure, with, or thunk
0x0007	S_SKIP	Skip - Reserve symbol space
0x0008	S_CVRESERVE	Reserved for CodeView internal use
0x0009	S_OBJNAME	Specify name of object file
0x000a	S_ENDARG	Specify end of arguments in function symbols
0x000b	S_COBOLUDT	Microfocus COBOL user-defined type
0x000c	S_MANYREG	Many register symbol
0x000d	S_RETURN	Function return description
0x000e	S_ENTRYTHIS	Description of this pointer at entry
0x0100	S_BPREL16	BP relative 16:16
0x0101	S_LDATA16	Local data 16:16
0x0102	S_GDATA16	Global data 16:16
0x0103	S_PUB16	Public symbol 16:16
0x0104	S_LPROC16	Local procedure start 16:16
0x0105	S_GPROC16	Global procedure start 16:16
0x0106	S_THUNK16	Thunk start 16:16
0x0107	S_BLOCK16	Block start 16:16
0x0108	S_WITH16	With start 16:16
0x0109	S_LABEL16	Code label 16:16
0x010a	S_CEXMODEL16	Change execution model 16:16
0x010b	S_VFTPATH16	Virtual function table path descriptor 16:16
0x010c	S_REGREL16	Specify 16:16 offset relative to arbitrary register.

0x0200	S_BPREL32	BP relative 16:32
0x0201	S_LDATA32	Local data 16:32
0x0202	S_GDATA32	Global data 16:32
0x0203	S_PUB32	Public symbol 16:32
0x0204	S_LPROC32	Local procedure start 16:32
0x0205	S_GPROC32	Global procedure start 16:32
0x0206	S_THUNK32	Thunk start 16:32
0x0207	S_BLOCK32	Block start 16:32
0x0208	S_WITH32	With start 16:32
0x0209	S_LABEL32	Label 16:32
0x020a	S_CEXMODEL32	Change execution model 16:32
0x020b	S_VFTPATH32	Virtual function table path descriptor 16:32
0x020c	S_REGREL32	16:32 offset relative to arbitrary register
0x020d	S_LTHREAD32	Local Thread Storage data
0x020e	S_GTHREAD32	Global Thread Storage data
0x0300	S_LPROCMIPS	Local procedure start MIPS
0x0301	S_GPROCMIPS	Global procedure start MIPS

0x0400	S_PROCREF	Reference to a procedure
0x0401	S_DATAREF	Reference to data
0x0402	S_ALIGN	Page align symbols
*/
int DumpCodeViewSymbols(PBYTE ptr, DWORD length,SEGMENT_DEBUG_INFO *sdi,int test)
{
    long           bytesToGo = length;
	int recordNr = 1;
	int recLen,idx;
	struct procInfo *pi;
	GENERIC_LIST *gl;
	unsigned char *start;

	if (test && 1 != *(unsigned long *)ptr) {
		HexDump(ptr,length);
		return 0;
	}
	start = ptr;
	if (test) {
		ptr += 4;
		bytesToGo -= 4;
	}
	while (bytesToGo > 0) {
		recLen = *(unsigned short *)(ptr);
		idx = *(unsigned short *)(ptr+2);
		printf("[%3d] offset %d length %d index %d (0x%XH) ",
				recordNr,(ptr-start),recLen,idx,idx);
		ptr += 4;
		switch (idx) {
			case S_REGISTER:
				DisplayRegister(ptr);
				break;
			case S_COMPILE:
				DisplayCompilerInfo(ptr);
				break;
			case S_UDT:
				DisplayUDT(ptr);
				break;
			case S_END:
				printf("Block End\n");
				break;
			case S_OBJNAME:
				DisplayObjName(ptr);
				break;
			case S_GPROC32:
			case S_LPROC32:
				pi = DisplayProcedureStart(idx,ptr);
				if (sdi) {
					gl = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST));
					gl->data = (char *)pi;
					if (idx == S_GPROC32) {
						gl->Next = sdi->GlobalProcedures;
						sdi->GlobalProcedures = gl;
					}
					else {
						gl->Next = sdi->LocalProcedures;
						sdi->LocalProcedures = gl;
					}
				}
				else free(pi);
				break;
			case S_LDATA32:
			case S_GDATA32:
			case S_PUB32:
				DisplayVariable(idx,ptr);
				break;
			case S_BPREL32:
				DisplayAutoVariable(ptr);
				break;
			case S_THUNK32:
				DisplayThunk32(ptr);
				break;
			case S_LABEL32:
				DisplayCodeLabel(ptr);
				break;
			case S_BLOCK32:
				DisplayBlockStart(ptr);
				break;
			case S_START:
				DisplayStartOfSearch(ptr);
				break;
			case S_PROCREF:
			case S_DATAREF:
				DisplayCVPack(ptr,idx);
				break;
			case S_ALIGN:
				printf("Pad record ");
				if (*(long *)ptr == -1) {
					printf("End");
				}
				printf("\n");
				break;
			default:
				printf("\n");
				if (recLen) {
					if (recLen < 500)
						HexDump(ptr,recLen);
					else {
						printf("***** reclen = %d!\n",recLen);
						return recordNr-1;
					}
				}
			break;
		}
		recordNr++;
		ptr += recLen - 2;
		bytesToGo -= 2 + recLen;
	}
	return (recordNr-1);
}
struct pdbInfo {
	unsigned short offset;
	unsigned short length1;
	unsigned short length2;
	unsigned short length3;
	char fname[1];
};
#define LF_MODIFIER  0x1
#define LF_POINTER   0x2
#define LF_STRUCTURE 0x5
#define LF_UNION     0x6
#define LF_PROCEDURE 0x8
#define LF_ARGLIST   0x201
#define LF_LIST      0x204
#define LF_MEMBER    0x406
#define LF_ARRAY     0x3
struct structInfo {
	short count;
	short field;
	short property;
	short dlist;
	short vshape;
	short length;
	char name[1];
};

struct pointerInfo {
	short attribute;
	short type;
	char variant[1];
};

struct procTypeInfo {
	short rvtype;
	char call;
	char reserved;
	short paramsCount;
	short arglistType;
};

struct structMemberInfo {
	short type;
	short attribute;
	short offset;
	char name[1];
};

struct structUnion {
	short count;
	short typeidx;
	short property;
	short length;
	char name[1];
};
#if 1
struct Types {
	short Type;
	char *name;
};

struct Types TypesTable[] = {

/* Special types */

{ 0x0000,	"no type"},
{ 0x0001,	"Absolute symbol"},
{ 0x0002,	"Segment type"},
{ 0x0003,	"Void"},
{ 0x0103,	"Near pointer to void"},
{ 0x0203,	"Far pointer to void"},
{ 0x0303,	"Huge pointer to void"},
{ 0x0403,	"32 bit near pointer to void"},
{ 0x0503,	"32 bit far pointer to void"},
{ 0x0004,	"Basic 8 byte currency value"},
{ 0x0005,	"Near Basic string"},
{ 0x0006,	"Far Basic string"},
{ 0x0007,	"Untranslated type record from CV 3.x format"},
{ 0x0060,	"Bit"},
{ 0x0061,	"Pascal CHAR"},

/* Character types */

{ 0x0010,	"8-bit signed"},
{ 0x0020,	"8-bit unsigned"},
{ 0x0110,	"Near pointer to 8-bit signed"},
{ 0x0120,	"Near pointer to 8-bit unsigned"},
{ 0x0210,	"Far pointer to 8-bit signed"},
{ 0x0220,	"Far pointer to 8-bit unsigned"},
{ 0x0310,	"Huge pointer to 8-bit signed"},
{ 0x0320,	"Huge pointer to 8-bit unsigned"},
{ 0x0410,	"16:32 near pointer to 8-bit signed"},
{ 0x0420,	"16:32 near pointer to 8-bit unsigned"},
{ 0x0510,	"16:32 far pointer to 8-bit signed"},
{ 0x0520,	"16:32 far pointer to 8-bit unsigned"},

/* Really a character types */

{ 0x0070,	"real char"},
{ 0x0170,	"near pointer to a real char"},
{ 0x0270,	"far pointer to a real char"},
{ 0x0370,	"huge pointer to a real char"},
{ 0x0470,	"16:32 near pointer to a real char"},
{ 0x0570,	"16:32 far pointer to a real char"},

/* Wide character types */

{ 0x0071,	"wide char"},
{ 0x0171,	"near pointer to a wide char"},
{ 0x0271,	"far pointer to a wide char"},
{ 0x0371,	"huge pointer to a wide char"},
{ 0x0471,	"16:32 near pointer to a wide char"},
{ 0x0571,	"16:32 far pointer to a wide char"},

/* Really 16 bit integer types */

{ 0x0072,	"16 bit signed int"},
{ 0x0073,	"16 bit unsigned int"},
{ 0x0172,	"near pointer to 16 bit signed int"},
{ 0x0173,	"near pointer to 16 bit unsigned int"},
{ 0x0272,	"far pointer to 16 bit signed int"},
{ 0x0273,	"far pointer to 16 bit unsigned int"},
{ 0x0372,	"huge pointer to 16 bit signed int"},
{ 0x0373,	"huge pointer to 16 bit unsigned int"},
{ 0x0472,	"16:32 near pointer to 16 bit signed int"},
{ 0x0473,	"16:32 near pointer to 16 bit unsigned int"},
{ 0x0572,	"16:32 far pointer to 16 bit signed int"},
{ 0x0573,	"16:32 far pointer to 16 bit unsigned int"},

/* 16-bit short types */

{ 0x0011,	"16-bit signed "},
{ 0x0021,	"16-bit unsigned"},
{ 0x0111,	"Near pointer to 16-bit signed"},
{ 0x0121,	"Near pointer to 16-bit unsigned"},
{ 0x0211,	"Far pointer to 16-bit signed"},
{ 0x0221,	"Far pointer to 16-bit unsigned"},
{ 0x0311,	"Huge pointer to 16-bit signed"},
{ 0x0321,	"Huge pointer to 16-bit unsigned"},
{ 0x0411,	"16:32 near pointer to 16 bit signed"},
{ 0x0421,	"16:32 near pointer to 16 bit unsigned"},
{ 0x0511,	"16:32 far pointer to 16 bit signed"},
{ 0x0521,	"16:32 far pointer to 16 bit unsigned"},

/* Really 32 bit integer types*/

{ 0x0074,	"32 bit signed int"},
{ 0x0075,	"32 bit unsigned int"},
{ 0x0174,	"near pointer to 32 bit signed int"},
{ 0x0175,	"near pointer to 32 bit unsigned int"},
{ 0x0274,	"far pointer to 32 bit signed int"},
{ 0x0275,	"far pointer to 32 bit unsigned int"},
{ 0x0374,	"huge pointer to 32 bit signed int"},
{ 0x0375,	"huge pointer to 32 bit unsigned int"},
{ 0x0474,	"16:32 near pointer to 32 bit signed int"},
{ 0x0475,	"16:32 near pointer to 32 bit unsigned int"},
{ 0x0574,	"16:32 far pointer to 32 bit signed int"},
{ 0x0575,	"16:32 far pointer to 32 bit unsigned int"},

/* 32-bit long types */

{ 0x0012,	"32-bit signed"},
{ 0x0022,	"32-bit unsigned"},
{ 0x0112,	"Near pointer to 32-bit signed"},
{ 0x0122,	"Near pointer to 32-bit unsigned"},
{ 0x0212,	"Far pointer to 32-bit signed"},
{ 0x0222,	"Far pointer to 32-bit unsigned"},
{ 0x0312,	"Huge pointer to 32-bit signed"},
{ 0x0322,	"Huge pointer to 32-bit unsigned"},
{ 0x0412,	"16:32 near pointer to 32 bit signed"},
{ 0x0422,	"16:32 near pointer to 32 bit unsigned"},
{ 0x0512,	"16:32 far pointer to 32 bit signed"},
{ 0x0522,	"16:32 far pointer to 32 bit unsigned"},

/* Really 64-bit int types */

{ 0x0076,	"64-bit signed  int"},
{ 0x0077,	"64-bit unsigned int"},
{ 0x0176,	"Near pointer to 64-bit signed int"},
{ 0x0177,	"Near pointer to 64-bit unsigned int"},
{ 0x0276,	"Far pointer to 64-bit signed int"},
{ 0x0277,	"Far pointer to 64-bit unsigned int"},
{ 0x0376,	"Huge pointer to 64-bit signed int"},
{ 0x0377,	"Huge pointer to 64-bit unsigned int"},
{ 0x0476,	"16:32 near pointer to 64 bit signed int"},
{ 0x0477,	"16:32 near pointer to 64 bit unsigned int"},
{ 0x0576,	"16:32 far pointer to 64 bit signed int"},
{ 0x0577,	"16:32 far pointer to 64 bit unsigned int"},

/* 64-bit integral types */

{ 0x0013,	"64-bit signed"},
{ 0x0023,	"64-bit unsigned"},
{ 0x0113,	"Near pointer to 64-bit signed"},
{ 0x0123,	"Near pointer to 64-bit unsigned"},
{ 0x0213,	"Far pointer to 64-bit signed"},
{ 0x0223,	"Far pointer to 64-bit unsigned"},
{ 0x0313,	"Huge pointer to 64-bit signed"},
{ 0x0323,	"Huge pointer to 64-bit unsigned"},
{ 0x0413,	"16:32 near pointer to 64 bit signed "},
{ 0x0423,	"16:32 near pointer to 64 bit unsigned "},
{ 0x0513,	"16:32 far pointer to 64 bit signed "},
{ 0x0523,	"16:32 far pointer to 64 bit unsigned "},

/* 32-bit real types */

{ 0x0040,	"32-bit real"},
{ 0x0140,	"Near pointer to 32-bit real"},
{ 0x0240,	"Far pointer to 32-bit real"},
{ 0x0340,	"Huge pointer to 32-bit real"},
{ 0x0440,	"16:32 near pointer to 32 bit real"},
{ 0x0540,	"16:32 far pointer to 32 bit real"},

/* 48-bit real types */

{ 0x0044,	"48-bit real"},
{ 0x0144,	"Near pointer to 48-bit real"},
{ 0x0244,	"Far pointer to 48-bit real"},
{ 0x0344,	"Huge pointer to 48-bit real"},
{ 0x0444,	"16:32 near pointer to 48 bit real"},
{ 0x0544,	"16:32 far pointer to 48 bit real"},

/* 64-bit real types */

{ 0x0041,	"64-bit real"},
{ 0x0141,	"Near pointer to 64-bit real"},
{ 0x0241,	"Far pointer to 64-bit real"},
{ 0x0341,	"Huge pointer to 64-bit real"},
{ 0x0441,	"16:32 near pointer to 64 bit real"},
{ 0x0541,	"16:32 far pointer to 64 bit real"},

/* 80-bit real types */

{ 0x0042,	"80-bit real"},
{ 0x0142,	"Near pointer to 80-bit real"},
{ 0x0242,	"Far pointer to 80-bit real"},
{ 0x0342,	"Huge pointer to 80-bit real"},
{ 0x0442,	"16:32 near pointer to 80 bit real"},
{ 0x0542,	"16:32 far pointer to 80 bit real"},

/* 128-bit real types */

{ 0x0043,	"128-bit real"},
{ 0x0143,	"Near pointer to 128-bit real"},
{ 0x0243,	"Far pointer to 128-bit real"},
{ 0x0343,	"Huge pointer to 128-bit real"},
{ 0x0443,	"16:32 near pointer to 128 bit real"},
{ 0x0543,	"16:32 far pointer to 128 bit real"},

/* 32-bit complex types */

{ 0x0050,	"32-bit complex"},
{ 0x0150,	"Near pointer to 32-bit complex"},
{ 0x0250,	"Far pointer to 32-bit complex"},
{ 0x0350,	"Huge pointer to 32-bit complex"},
{ 0x0450,	"16:32 near pointer to 32 bit complex"},
{ 0x0550,	"16:32 far pointer to 32 bit complex"},

/* 64-bit complex types */

{ 0x0051,	"64-bit complex"},
{ 0x0151,	"Near pointer to 64-bit complex"},
{ 0x0251,	"Far pointer to 64-bit complex"},
{ 0x0351,	"Huge pointer to 64-bit complex"},
{ 0x0451,	"16:32 near pointer to 64 bit complex"},
{ 0x0551,	"16:32 far pointer to 64 bit complex"},

/* 80-bit complex types */

{ 0x0052,	"80-bit complex"},
{ 0x0152,	"Near pointer to 80-bit complex"},
{ 0x0252,	"Far pointer to 80-bit complex"},
{ 0x0352,	"Huge pointer to 80-bit complex"},
{ 0x0452,	"16:32 near pointer to 80 bit complex"},
{ 0x0552,	"16:32 far pointer to 80 bit complex"},

/* 128-bit complex types */

{ 0x0053,	"128-bit complex"},
{ 0x0153,	"Near pointer to 128-bit complex"},
{ 0x0253,	"Far pointer to 128-bit complex"},
{ 0x0353,	"Huge pointer to 128-bit real"},
{ 0x0453,	"16:32 near pointer to 128 bit complex"},
{ 0x0553,	"16:32 far pointer to 128 bit complex"},

/* Boolean types */

{ 0x0030,	"8-bit Boolean"},
{ 0x0031,	"16-bit Boolean"},
{ 0x0032,	"32-bit Boolean"},
{ 0x0033,	"64-bit Boolean"},
{ 0x0130,	"Near pointer to  8-bit Boolean"},
{ 0x0131,	"Near pointer to 16-bit Boolean"},
{ 0x0132,	"Near pointer to 32-bit Boolean"},
{ 0x0133,	"Near pointer to 64-bit Boolean"},
{ 0x0230,	"Far pointer to  8-bit Boolean"},
{ 0x0231,	"Far pointer to 16-bit Boolean"},
{ 0x0232,	"Far pointer to 32-bit Boolean"},
{ 0x0233,	"Far pointer to 64-bit Boolean"},
{ 0x0330,	"Huge pointer to 8-bit Boolean"},
{ 0x0331,	"Huge pointer to 16-bit Boolean"},
{ 0x0332,	"Huge pointer to 32-bit Boolean"},
{ 0x0333,	"Huge pointer to 64-bit Boolean"},
{ 0x0430,	"16:32 near pointer to 8 bit Boolean"},
{ 0x0530,	"16:32 far pointer to 8 bit Boolean"},
{ 0x0431,	"16:32 near pointer to 16 bit Boolean"},
{ 0x0531,	"16:32 far pointer to 16 bit Boolean"},
{ 0x0432,	"16:32 near pointer to 32 bit Boolean"},
{ 0x0532,	"16:32 far pointer to 32 bit Boolean"},
{ 0x0433,	"16:32 near pointer to 64-bit Boolean"},
{ 0x0533,	"16:32 far pointer to 64-bit Boolean"},
{ 0,	NULL					}
};
#endif
static char TypeName[80];

char *DisplayType(int type)
{
	int i = 0;
	while (TypesTable[i].name) {
		if (TypesTable[i].Type == type) {
			strcpy(TypeName,TypesTable[i].name);
			return TypeName;
		}
		i++;
	}
	sprintf(TypeName,"0x%x (%d %d)",type,type,type-4096);
	return(TypeName);
}
/*
(0x0006) Unions
The format for a Union is as follows:
2			2		2		2			*		*
LF_UNION	count	@field	property	length	name
count 	Number of fields in the union
@field 	Type index of field  list.
property	Property bit field.
length	Numeric leaf specifying size in bytes of the union
name	Length-prefixed name of union
*/
void DisplayUnionType(char *data)
{
	struct structUnion *s;

	s = (struct structUnion *)data;
	printf("Union '");
	DisplayLengthPrefixedString(s->name);
	printf("'\n%snr of flds %d type idx %d property %d length %d\n",
			indent,s->count,s->typeidx,s->property,s->length);
}
/*
(0x0008) Procedure
2				2		1		1			2		2
LF_PROCEDURE	@rvtype	call	reserved	#parms	@arglist
@rvtype	Type index of the value returned by the procedure
call	Calling convention of the procedure
0	Near C (arguments pushed right to left, caller pops arguments)
1	Far C
2	Near pascal  (arguments pushed left to right, callee pops arguments)
3	Far pascal
4	Near fastcall
5	Far fastcall
6	Reserved
7	Near stdcall
8	Far stdcall
9	Near syscall
10	Far syscall
11	This call
12	MIPS call
13 	Generic
14-255	Reserved
#parms	Number of parameters
@arglist	Type index of argument list type record

*/
void DisplayProcTypeInfo(char *data)
{
	struct procTypeInfo *s;
	int ti;

	s = (struct procTypeInfo *)data;
	ti = s->arglistType;
	printf("Procedure:\n%sreturn type: %d call %d parameters %d arglist idx %d (%d)\n",
			indent,s->rvtype,s->call,s->paramsCount,ti,(ti)? ti - 4096:0);
}

void DisplayTypeModifier(char *data,int len)
{
	printf("Type modifier: attribute:0x%x @index %d\n",
		*(unsigned short *)data,*(unsigned short *)(data+2));
}
void DisplayPointerInfo(char *data,int len)
{
	struct pointerInfo *s;
	int i;

	s = (struct pointerInfo *)data;
	printf("Pointer: attribute 0x%X type %d ",s->attribute,s->type);
	if (s->type >= 4096) printf("(%d) ",s->type-4096);
	len -= 4;
	for (i=0; i<len;i++) {
		printf("%02X ",s->variant[i]);
	}
	printf("\n");
}
/*
The format for types and classes is as follows:
2		2		2		2			2		2		*		*
leaf	count	@field	property	@dList	@vshape	length	name
leaf	LF_CLASS or LF_STRUCTURE
count	Number of elements  in the class or structure.
		This count includes direct, virtual, and indirect virtual bases, and methods
		including overloads, data members, static data members, friends, and so on.
@field	Type index of the field list for this class.
property	Property bit field
packed	:1	Structure is packed
ctor	:1	Class has constructors and/or destructors
overops	:1	Class has overloaded operators
isnested	:1	Class is a nested class
cnested	:1	Class contains nested classes
opassign	:1	Class has overloaded assignment
opcast	:1	Class has casting methods
fwdref	:1	Class/structure is a forward (incomplete) reference
scoped	:1	This is a scoped definition
reserved	:8
@dList	Type index of the derivation list.
		This is output by the compiler as 0x0000 and is filled in by the CVPACK
		utility to a  LF_DERIVED record containing the type indices of those
		classes which immediately inherit the current class.
		A zero index indicates that no derivation information is available.
		A LF_NULL index indicates that  the class is not inherited by other classes.
@vshape	Type index of the virtual function table shape descriptor
length	Numeric leaf specifying size in bytes of the structure
name	Length-prefixed name this type
*/
void DisplayStructInfo(char *data)
{
	struct structInfo *s;
	int f;

	s = (struct structInfo *)data;
	f = s->field - 4096;
	printf("Structure '");
	DisplayLengthPrefixedString(s->name);
	printf("'\n%selements %d field list %d (%d) property %d dlist %d vshape %d bytes %d\n",
			indent,s->count,s->field,f,s->property,s->dlist,s->vshape,s->length);
}
/*
(0x0201) Argument List
2			2			*
LF_ARGLIST	argcount	indices
argcount	Count of number of indices in list
indices 	List of type indices for describing the formal parameters to a function
			or method.
*/
void DisplayArgList(char *data)
{
	unsigned short *s;
	int c,i;

	s = (unsigned short *)data;
	c = *s++;
	printf("Argument list. %d arguments\n",c);
	if (c > 255) {
		printf("error: bad synch\n");
		return;
	}
	for (i=0; i<c;i++) {
		printf("%sArgument %d type %d (0x%XH)\n",indent,i+1,*s,*s);
		s++;
	}

}
/*
(0x0406) Data Member
This leaf specifies nonstatic data members of a class.
2		2	2		*			*
LF_MEMBER	@type	attribute	offset	name
@type	Index to type record for field
attribute	Member attribute bit field
offset 	Numeric leaf specifying the offset of field in the structure
name	Length-prefixed name of the member field
*/
int DisplayStructMember(char *data)
{
	struct structMemberInfo *s;

	s = (struct structMemberInfo *)data;
	printf("%sMember '",indent);
	DisplayLengthPrefixedString(s->name);
	printf("' type %s attribute 0x%X offset %d\n",DisplayType(s->type),s->attribute,s->offset);
	return(7 + s->name[0]);
}

void DumpListTypes(char *data,int len)
{
    long           bytesToGo = len;
	int recLen,idx;

	printf("Structure member list\n");
	while (bytesToGo > 0) {
		idx = *(unsigned short *)(data);
		data += 2;
		bytesToGo -= 2;
		switch(idx) {
			case LF_MEMBER:
				recLen = DisplayStructMember(data);
				bytesToGo -= recLen;
				data += recLen;
				if ((unsigned int)*data > (unsigned int)0xF0) {
					bytesToGo -= (*data & 0x0F);
					data += (*data & 0x0F);
				}
				break;
			default:
				HexDump(data,bytesToGo);
				bytesToGo = 0;
				break;
		}
	}
}

static char *NumericLeaf(char *data,int *result)
{
	unsigned short s;

	s = *(unsigned short *)data;
	data += 2;
	if (s > 0x8000) {
		*result = *(unsigned long *)data;
		data += 4;
		return(data);
	}
	else {
		*result = s;
		return(data);
	}
}

/*
0x0003) Simple Array

2			2			2			*		*

LF_ARRAY	@elemtype	@idxtype	length	name

@elemtype	Type index of each array element
@idxtype	Type index of indexing variable
length	Length of array in bytes
name	Length-prefixed name of array
*/

void DisplaySimpleArray(char *data)
{
	short elemtype,idxtype;
	int length;
	char buf[120],*p;

	elemtype = *(unsigned short *)data;
	idxtype = *(unsigned short *)(data +2);
	p = NumericLeaf(data + 4,&length);
	printf("'");
	DisplayLengthPrefixedString(p);
	strcpy(buf,DisplayType(elemtype));
	printf("'simple array type %s (0x%x)\n%sidxtype %s (%d) len %d\n",buf,elemtype,indent,
		DisplayType(idxtype),idxtype,length);
}

void DumpCodeViewTypes(PBYTE ptr, DWORD length,int checkSignature,long *lp)
{
    long           bytesToGo = length;
	int recordNr = 0,tabidx=0;
	int recLen,idx,dofree,offset;
	unsigned char *start;
	struct pdbInfo *s;

	if (checkSignature) {
		if (1 != *(unsigned long *)ptr) {
			HexDump(ptr,length);
			return;
		}
		ptr += 4;
		bytesToGo -= 4;
	}
	dofree = 0;
	start = ptr;

	while (bytesToGo > 0) {
		recLen = *(unsigned short *)(ptr);
		idx = *(unsigned short *)(ptr+2);
		offset = (int)(ptr - start);
		ptr += 4;
		printf("[%3d] Offset %d length %d index %d (0x%XH) ",recordNr,offset,recLen,idx,idx);
		if (recLen > 3000) {
			printf("Record length > 3000. Quitting.\n");
			return;
		}
		switch (idx) {
			case 0x16:
				s = (struct pdbInfo *)ptr;
				printf("%d 0x%X %d 0x%X %d 0x%X\n%s",
					s->offset,s->offset,
					s->length1,s->length1,
					s->length2,s->length2,indent);
				DisplayLengthPrefixedString(s->fname);
				printf("\n");
#if 0
				p = s->fname;
				c = *p++;
				name = malloc(c+1);
				memset(name,0,c+1);
				for (i=0; i<c;i++) name[i] = *p++;
				f = fopen(name,"rb");
				free(name);
				if (f == NULL) break;
				ptr = malloc(s->length1);
				length = s->length1;
				fseek(f,s->offset,SEEK_SET);
				fread(ptr,1,length,f);
				fclose(f);
				dofree = 1;
				bytesToGo = length;
				printf("\n");
				goto restart;
#else
				HexDump(ptr,length);
				break;
#endif
			case LF_STRUCTURE:
				DisplayStructInfo(ptr);
				break;
			case LF_UNION:
				DisplayUnionType(ptr);
				break;
			case LF_MODIFIER:
				DisplayTypeModifier(ptr,recLen-2);
				break;
			case LF_POINTER:
				DisplayPointerInfo(ptr,recLen-2);
				break;
			case LF_PROCEDURE:
				DisplayProcTypeInfo(ptr);
				break;
			case LF_ARGLIST:
				DisplayArgList(ptr);
				break;
			case 0x204:
				DumpListTypes(ptr,recLen-2);
				break;
			case LF_ARRAY:
				DisplaySimpleArray(ptr);
				break;
			default:
				printf("\n");
				if (recLen) HexDump(ptr,recLen-2);
		}
		recordNr++;
		if (lp) {
			ptr += lp[tabidx]-4;
			bytesToGo -= lp[tabidx];
			tabidx++;
		}
		else {
			ptr += recLen - 2;
			bytesToGo -= 2 + recLen;
		}
	}
	if (dofree) free(ptr);
}

int FindRelocation(unsigned long addr, char *name, PIMAGE_SECTION_HEADER pSection, int range)
{
    int             k, count, idx;
    PIMAGE_RELOCATION pRelocs;

	if (pSection == NULL) return 0;
    count = pSection->NumberOfRelocations;
    pRelocs = MakePtr(PIMAGE_RELOCATION, pImageFileHeader, pSection->PointerToRelocations);
    for (k = 0; k < count; k++) {
        /* if (pRelocs->VirtualAddress > addr) break; */
		if (pRelocs->VirtualAddress >= addr && pRelocs->VirtualAddress <= (addr+range)) {
            idx = pRelocs->SymbolTableIndex;
			if ((unsigned)idx >= COFFSymbolCount) {
				sprintf(name,"Bad sym index %d (addr %ld)",idx,addr);
			}
			else if (pSymbolTable[idx].N.Name.Short != 0) {
                sprintf(name, "%s", pSymbolTable[idx].N.ShortName);
                name[8] = 0;
                return (1);
            }
            else
                sprintf(name, "%s", stringTable + pSymbolTable[idx].N.Name.Long);

            return (1);
        }
        pRelocs++;
    }
    return (0);
}
int             printAddr(unsigned long addr, char *where)
{
    unsigned char   name[64];

    if (FindRelocation(addr - 4, name, pCodeSection, 0)) {
        sprintf(where, "%s", name);
    }
    else if (FindRelocation(addr - 4, name, pDataSection, 0)) {
        sprintf(where, "%s", name);
    }
    else
        sprintf(where, "%ld", addr);
    return (strlen(where));
}

int FindRelocationInInstruction(int addr,int range,char *where)
{
    unsigned char   name[256];
    if (FindRelocation(addr, name, pCodeSection, range)) {
        sprintf(where, "%s", name);
    }
    else if (FindRelocation(addr, name, pDataSection, range)) {
        sprintf(where, "%s", name);
    }
    else
        *where = 0;
    return (strlen(where));

}
int isFunction(int pc,char *outp)
{
	int i,j;
	char *name,*p,ss[9];
	PIMAGE_AUX_SYMBOL auxSym;
	PIMAGE_SYMBOL pSymbol;

	for (i=0; i<NumberOfFunctions;i++) {
		if (FunctionIndexTable[i]->Value == (unsigned)pc) {
			pSymbol = FunctionIndexTable[i];
		    auxSym = (PIMAGE_AUX_SYMBOL) (pSymbol + 1);
	        if (pSymbol->N.Name.Short != 0) {
				p = pSymbol->N.ShortName;
				for (j=0; j<8; j++) {
					ss[j] = p[j];
				}
				ss[8] = 0;
				name = ss;
			}
    	    else {
				name = stringTable + pSymbol->N.Name.Long;
			}
			sprintf(outp,"%s:   Size %5ld",name,auxSym->Sym.Misc.TotalSize);
			return(1);
		}
	}
	return(0);
}

int FindLineNr(int idx)
{
	int i;

	for (i=0; i<ObjLineNumberIdx;i++) {
		if (LineNumberTable[i].Offset == idx) return(LineNumberTable[i].LnNumber);
	}
	return(FindExeLineNrs(idx));
}

static char *fbuffer = NULL;
static char **LinePointers;
static int LastLineRead = 0;
static char *lastP;

int FindLines(char *buf,int size)
{
	int i,lines,j;

	lines = 0;
	for (i=0; i<size;i++) {
		if (buf[i] == '\n') lines++;
	}
	LinePointers = (char **)malloc((lines+1) * sizeof(char *));
	LinePointers[0] = buf;
	j = 1;
	for (i=0; i<size;i++) {
		if (buf[i] == '\n') {
			LinePointers[j++] = &buf[i];
		}
	}
	return(lines);
}

int ReadFileIn(char *name)
{
	FILE *f;
	long size;

	f = fopen(name,"r");
	if (f == NULL) return(0);
	fseek(f,0,SEEK_END);
	size = ftell(f);
	fseek(f,0,SEEK_SET);
	fbuffer = malloc(size+5);
	if (fbuffer == NULL) return(0);
	fread(fbuffer,1,size,f);
	fclose(f);
	fbuffer[size] = '\n';
	fbuffer[size+1] = 0;
	return(FindLines(fbuffer,size));
}

static int nofile = 0;
int ReadNthLine(int n,char *buf)
{
	char *endP;
	int line,j;

	*buf = 0;
	if (nofile) return 0;
	if (LinePointers == NULL) {
		nofile = 1;
		return 0;
	}
	j = 0;
	if (fbuffer == NULL) {
		if (*SourceFileName == 0) return(0);
		if (!ReadFileIn(SourceFileName)) {
			nofile = 1;
			return 0;
		}
		LastLineRead = 0;
	}
	line = LastLineRead;
	endP = LinePointers[n];
	lastP = LinePointers[LastLineRead];
	while (lastP < endP) {
		if (*lastP == '\n') {
			sprintf(buf+j,"\n[%4d] ",line+1);
			line++;
			j += 8;
			lastP++;
		}
		else {
			buf[j++] = *lastP++;
		}
	}
	buf[j++] = '\n';
	buf[j] = 0;
	LastLineRead = n;

	return(j);
}


