#include <windows.h>
#ifdef WINELIB
#include <unix.h>
#include <peexe.h>
#endif
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "objdump.h"
#include "exedump.h"
#include "extrnvar.h"

// Global variables set here, and used in EXEDUMP.C and OBJDUMP.C
BOOL fShowRelocations = FALSE;
BOOL fShowRawSectionData = FALSE;
BOOL fShowSymbolTable = FALSE;
BOOL fShowLineNumbers = FALSE;
BOOL fExtract = FALSE;
BOOL fResources = FALSE;
BOOL fDebug = FALSE;
int TotalFileSize = 0;
char HelpText[] =
"\nPEDUMP - Win32/COFF .EXE/.OBJ file dumper - 1993 Matt Pietrek, 1995-1997 Jacob Navia\n"
"Working version "__DATE__" " __TIME__"\n\n"
"Syntax: PEDUMP [switches] filename\n\n"
#ifdef WINELIB 
"  -a    include everything in dump\n"
"  -h    include dump of sections\n"
"  -l    include line number information\n"
"  -rel  show base relocations\n"
"  -e    show resources\n"
"  -x    eXtract section: for instance /x.data\n"
"  -d    dump code view format debug info\n"
"  -s    show symbol table\n";
#else
"  /A    include everything in dump\n"
"  /H    include dump of sections\n"
"  /L    include line number information\n"
"  /Rel  show base relocations\n"
"  /E    show resources\n"
"  /X    eXtract section: for instance /x.data\n"
"  /D    dump code view format debug info\n"
"  /S    show symbol table\n";
#endif
int stricmp (const char *s1,const char *s2)
{
  register  unsigned char *p1 = ( unsigned char *) s1;
  register  unsigned char *p2 = ( unsigned char *) s2;
  register int ret;
  unsigned char c1;

  if (p1 == p2)
    return 0;

  for (; !(ret = (c1 = tolower(*p1)) - tolower(*p2)); p1++, p2++)
    if (c1 == '\0') break;
  return ret;
}

//
// Open up a file, memory map it, and call the appropriate dumping routine
//
LPVOID lpFileBase;
void DumpFile(LPSTR filename)
{
	HANDLE hFile;
	HANDLE hFileMapping;
	PIMAGE_DOS_HEADER dosHeader;

	hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL,
						OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		printf("Couldn't open file with CreateFile()\n");
		return;
	}
	TotalFileSize = SetFilePointer(hFile,0,NULL,FILE_END);
	SetFilePointer(hFile,0,NULL,FILE_BEGIN);

	hFileMapping = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if ( hFileMapping == 0 )
	{
		CloseHandle(hFile);
		printf("Couldn't open file mapping with CreateFileMapping()\n");
		return;
	}

	lpFileBase = MapViewOfFile(hFileMapping, FILE_MAP_READ, 0, 0, 0);
	if ( lpFileBase == 0 )
	{
		CloseHandle(hFileMapping);
		CloseHandle(hFile);
		printf("Couldn't map view of file with MapViewOfFile()\n");
		return;
	}

	printf("Dump of file %s (%d bytes)\n\n", filename,TotalFileSize);


	dosHeader = (PIMAGE_DOS_HEADER)lpFileBase;
	if ( dosHeader->e_magic == IMAGE_DOS_SIGNATURE )
	{
		DumpExeFile( dosHeader );
	}
	else if ( (dosHeader->e_magic == 0x014C)	// Does it look like a i386
		      && (dosHeader->e_sp == 0) )		// COFF OBJ file???
	{
		// The two tests above aren't what they look like.  They're
		// really checking for IMAGE_FILE_HEADER.Machine == i386 (0x14C)
		// and IMAGE_FILE_HEADER.SizeOfOptionalHeader == 0;

		DumpObjFile( (PIMAGE_FILE_HEADER)lpFileBase );
	}
	else if (!strncmp("!<arch>\n",(char *)lpFileBase,8)) {
		DumpLibFile((char *)lpFileBase);
	}
	else if (dosHeader->e_sp == 0 && !strncmp((unsigned char *)lpFileBase + 20,".rsrc$",6))
		DumpObjFile((PIMAGE_FILE_HEADER)lpFileBase);
	else
		printf("unrecognized file format\n");
	UnmapViewOfFile(lpFileBase);
	CloseHandle(hFileMapping);
	CloseHandle(hFile);
}
void UpperCase(char *s)
{
	while (*s) {
		if (*s >= 'a' && *s <= 'z') *s -= ' ';
		s++;
	}
}
char ExtractSectionName[20];
//
// process all the command line arguments and return a pointer to
// the filename argument.
//
PSTR ProcessCommandLine(int argc, char *argv[])
{
	int i;

	for ( i=1; i < argc; i++ )
	{
#ifdef  WINELIB
		if (argv[i][0] == '-')
		{
                        UpperCase(argv[i]);
#else
		UpperCase(argv[i]);
		// Is it a switch character?
		if ( (argv[i][0] == '-') || (argv[i][0] == '/') )
		{
#endif
			if ( argv[i][1] == 'A' )
			{
				fShowRelocations = TRUE;
				fShowRawSectionData = TRUE;
				fShowSymbolTable = TRUE;
				fShowLineNumbers = TRUE;
				fDebug = TRUE;
			}
			else if ( argv[i][1] == 'H' )
				fShowRawSectionData = TRUE;
			else if ( argv[i][1] == 'L' )
				fShowLineNumbers = TRUE;
			else if ( argv[i][1] == 'R' )
				fShowRelocations = TRUE;
			else if (argv[i][1] == 'X') {
				char *p = argv[i] + 2,*dst = ExtractSectionName;

				while (*p) {
					if (*p >= 'A' && *p <= 'Z') *p += ' ';
					*dst++ = *p++;
				}
				fExtract = TRUE;
			}
			else if (argv[i][1] == 'D') {
				fDebug = 1;
			}
			else if (argv[i][1] == 'E') {
				fResources = 1;
			}
			else if ( argv[i][1] == 'S' )
				fShowSymbolTable = TRUE;
		}
		else	// Not a switch character.  Must be the filename
		{
			break;
		}
	}
	return argv[i];
}

char *dargv[] = { "pedump.exe", "/D", "d:\\nedit1\\win\\nlcc\\wedit.exe"};
int dargc = 3;
int main(int argc, char *argv[])
{
	PSTR filename;

	if ( argc == 1 )
	{
		printf(	HelpText );
		return 1;
	}

	filename = ProcessCommandLine(argc, argv);
	if ( filename )
		DumpFile( filename );
	else
	{
		printf(	HelpText );
		return 1;
	}
	return 0;
}
int IsGoodPointer(void *p)
{
	if (((DWORD)p >= (DWORD)lpFileBase) && ((DWORD)p <= (DWORD)lpFileBase + TotalFileSize))
		return(1);
	printf("*****  Bad pointer 0x$%x (%d) ***** \n",(unsigned int)p,(int)p);
	return(0);
}
